import csv
from typing import List, Generator, Any, Dict, IO

import pandas as pd

from .dataset import Dataset
from ..types import RecordType


class CsvDataset(Dataset[IO[str]]):
    READ_ARGS: Dict[str, Any] = {
        "mode": "r",
        "encoding": "utf-8",
        "errors": "ignore",
    }

    def count_rows(self) -> int:
        with self.fileobj() as f:
            reader = csv.reader(f)

            # handle case where CSV is empty
            return max(
                len(list(reader)) - 1,  # first row is headers
                0,
            )

    def get_columns(self) -> List[str]:
        with self.fileobj() as f:
            reader = csv.reader(f)
            columns = next(reader)
            return columns

    def read_streaming_records(self) -> Generator[RecordType, None, None]:
        with self.fileobj() as f:
            reader = csv.reader(f)
            columns = next(reader)
            for row in reader:
                yield dict(zip(columns, row))

    def read_streaming_values(self) -> Generator[List[Any], None, None]:
        with self.fileobj() as f:
            reader = csv.reader(f)
            next(reader)
            for row in reader:
                yield row

    def read_file_as_dataframe(self) -> pd.DataFrame:
        with self.fileobj() as f:
            return pd.read_csv(f, keep_default_na=True)



from .dataset_source import DatasetSource as DatasetSource
from .dataframe_dataset_source import DataFrameDatasetSource as DataFrameDatasetSource
from .filepath_dataset_source import FilepathDatasetSource as FilepathDatasetSource
from .pandas_dataset_source import PandasDatasetSource as PandasDatasetSource

try:
    import pyspark.sql
    from .pyspark_dataset_source import PySparkDatasetSource as PySparkDatasetSource
except ImportError:
    pass



class InvalidDatasetError(ValueError):
    pass


class EmptyDatasetError(InvalidDatasetError):
    pass


class ColumnMismatchError(InvalidDatasetError):
    pass


class APIError(Exception):
    pass


class UnsupportedVersionError(APIError):
    def __init__(self) -> None:
        super().__init__(
            "cleanlab-studio is out of date and must be upgraded. Run 'pip install --upgrade cleanlab-studio'."
        )


class AuthError(APIError):
    def __init__(self) -> None:
        super().__init__("invalid API key")


class InternalError(Exception):
    pass


class CleansetError(InternalError):
    pass



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Set, Optional, Any

from cleanlab_studio.cli.types import Modality

SchemaMetadataDictType = Dict[str, Optional[str]]
SchemaFieldsDictType = Dict[str, Dict[str, str]]


class DataType(Enum):
    string = "string"
    integer = "integer"
    float = "float"
    boolean = "boolean"


class FeatureType(Enum):
    identifier = "identifier"
    categorical = "categorical"
    numeric = "numeric"
    text = "text"
    boolean = "boolean"
    datetime = "datetime"
    image = "image"


DATA_TYPES_TO_FEATURE_TYPES: Dict[DataType, Set[FeatureType]] = {
    DataType.string: {
        FeatureType.text,
        FeatureType.categorical,
        FeatureType.datetime,
        FeatureType.identifier,
        FeatureType.image,
    },
    DataType.integer: {
        FeatureType.categorical,
        FeatureType.datetime,
        FeatureType.identifier,
        FeatureType.numeric,
    },
    DataType.float: {FeatureType.datetime, FeatureType.numeric},
    DataType.boolean: {FeatureType.boolean},
}


MEDIA_FEATURE_TYPES: List[FeatureType] = [FeatureType.image]


@dataclass
class FieldSpecification:
    data_type: DataType
    feature_type: FeatureType

    @staticmethod
    def create(data_type: str, feature_type: str) -> FieldSpecification:
        data_type_ = DataType(data_type)
        feature_type_ = FeatureType(feature_type)

        if feature_type_ not in DATA_TYPES_TO_FEATURE_TYPES[data_type_]:
            raise ValueError(
                f"Invalid column feature type: '{feature_type_.value}' for data type: '{data_type_.value}'. "
                f"Accepted categories for type '{data_type_.value}' are: {', '.join(t.value for t in DATA_TYPES_TO_FEATURE_TYPES[data_type_])}"
            )
        return FieldSpecification(data_type=data_type_, feature_type=feature_type_)

    def to_dict(self) -> Dict[str, str]:
        return dict(data_type=self.data_type.value, feature_type=self.feature_type.value)


@dataclass
class SchemaMetadata:
    id_column: str
    modality: Modality
    name: str

    @staticmethod
    def create(id_column: str, modality: str, name: str) -> SchemaMetadata:
        return SchemaMetadata(
            id_column=id_column,
            modality=Modality(modality),
            name=name,
        )

    def to_dict(self) -> SchemaMetadataDictType:
        return dict(
            id_column=self.id_column,
            modality=self.modality.value,
            name=self.name,
        )


@dataclass
class Schema:
    metadata: SchemaMetadata
    fields: Dict[str, FieldSpecification]
    version: str

    @staticmethod
    def create(
        metadata: SchemaMetadataDictType, fields: Dict[str, Dict[str, str]], version: str
    ) -> Schema:
        fields_ = dict()
        for field, field_spec in fields.items():
            if not isinstance(field, str):
                raise ValueError(
                    f"All schema columns must be strings. Found non-string column: {field}"
                )
            if field == "":
                raise ValueError(
                    "Found empty string for schema column name. Schema columns cannot be empty strings."
                )
            fields_[field] = FieldSpecification.create(
                data_type=field_spec["data_type"], feature_type=field_spec["feature_type"]
            )

        # metadata variables
        id_column = metadata["id_column"]
        modality = metadata["modality"]
        name = metadata["name"]
        assert id_column is not None
        assert modality is not None
        assert name is not None
        return Schema(
            metadata=SchemaMetadata.create(
                id_column=id_column,
                modality=modality,
                name=name,
            ),
            fields=fields_,
            version=version,
        )

    def to_dict(self) -> Dict[str, Any]:
        retval: Dict[str, Any] = dict()
        retval["metadata"] = self.metadata.to_dict()
        fields: Dict[str, Dict[str, str]] = dict()
        for field, field_spec in self.fields.items():
            fields[field] = field_spec.to_dict()
        retval["fields"] = fields
        retval["version"] = self.version

        return retval

    @classmethod
    def from_dict(cls, schema_dict: Dict[str, Any]) -> Schema:
        return Schema.create(
            metadata=schema_dict["metadata"],
            fields=schema_dict["fields"],
            version=schema_dict["version"],
        )


PYTHON_TYPES_TO_READABLE_STRING: Dict[type, str] = {
    str: DataType.string.value,
    float: DataType.float.value,
    int: DataType.integer.value,
    bool: DataType.boolean.value,
}

DATA_TYPES_TO_PYTHON_TYPES: Dict[DataType, type] = {
    DataType.string: str,
    DataType.float: float,
    DataType.integer: int,
    DataType.boolean: bool,
}



import pathlib
from typing import List

import pandas as pd
import pytest

from cleanlab_studio.cli.classes.csv_dataset import CsvDataset
from .constants import DATASETS_DIR
from .utils import assert_dataset_matches_dataframe, csv_filepath_to_dataframe


CSV_DATASETS: List[pathlib.Path] = [*DATASETS_DIR.glob(r"*.csv")]


@pytest.mark.parametrize(
    ("csv_filepath", "csv_df"),
    ((csv_filepath, csv_filepath_to_dataframe(csv_filepath)) for csv_filepath in CSV_DATASETS),
)
def test_csv_dataset_from_filepath_matches_expected(
    csv_filepath: pathlib.Path, csv_df: pd.DataFrame
):
    """Tests that CsvDataset, loaded from filepath, matches CSV loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load CsvDataset from filepath
    csv_dataset = CsvDataset(filepath=str(csv_filepath))

    # check that methods match values read by pandas
    assert_dataset_matches_dataframe(csv_dataset, csv_df)


@pytest.mark.parametrize(
    ("csv_filepath", "csv_df"),
    ((csv_filepath, csv_filepath_to_dataframe(csv_filepath)) for csv_filepath in CSV_DATASETS),
)
def test_csv_dataset_from_fileobj_matches_expected(
    csv_filepath: pathlib.Path, csv_df: pd.DataFrame
):
    """Tests that CsvDataset, loaded from file object, matches CSV loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load CsvDataset from object
    with open(csv_filepath, **CsvDataset.READ_ARGS) as csv_file:
        csv_dataset = CsvDataset(fileobj=csv_file)

        # check that methods match values read by pandas
        assert_dataset_matches_dataframe(csv_dataset, csv_df)



import datetime
from typing import List, Generator, Any, Dict, IO

from cleanlab_studio.cli.classes.dataset import Dataset
import pyexcel
import pandas as pd

from cleanlab_studio.cli.types import RecordType


class ExcelDataset(Dataset[IO[bytes]]):
    READ_ARGS: Dict[str, str] = {"mode": "rb"}

    def __init__(self, *args: Any, file_type: str, **kwargs: Any):
        self._file_type = file_type
        super().__init__(*args, **kwargs)

    def read_streaming_records(self) -> Generator[RecordType, None, None]:
        with self.fileobj() as f:
            for r in pyexcel.iget_records(file_stream=f, file_type=self._file_type):
                yield self._preprocess_record(r)

    def read_streaming_values(self) -> Generator[List[Any], None, None]:
        with self.fileobj() as f:
            stream = pyexcel.iget_array(file_stream=f, file_type=self._file_type)
            next(stream)  # skip header row
            for r in stream:
                yield self._preprocess_values(r)

    def read_file_as_dataframe(self) -> pd.DataFrame:
        with self.fileobj() as f:
            return pd.read_excel(f, keep_default_na=True)

    def get_columns(self) -> List[str]:
        with self.fileobj() as f:
            stream = pyexcel.iget_array(file_stream=f, file_type=self._file_type)

        return [str(col) for col in next(stream)]

    def _preprocess_record(self, record: RecordType) -> RecordType:
        """Preprocesses record.

        :param record: record to preprocess
        :return: preprocess record
        """
        return {
            record_key: self._preprocess_value(record_value)
            for record_key, record_value in record.items()
        }

    def _preprocess_values(self, values: List[Any]) -> List[Any]:
        """Preprocesses values.

        :param values: values to preprocess
        :return: preprocess values
        """
        return [self._preprocess_value(value) for value in values]

    def _preprocess_value(self, record_value: Any) -> Any:
        """Preprocesses record value.
        Operations performed:
            - Cast datetimes to strings

        :param record_value: record value to preprocess
        :return: preprocessed record value
        """
        if isinstance(
            record_value, (datetime.date, datetime.time, datetime.datetime, pd.Timestamp)
        ):
            return record_value.isoformat()

        return record_value



import asyncio
import functools
import json
from typing import List, Optional
from tqdm import tqdm

import aiohttp
from multidict import CIMultiDictProxy
import requests

from .api import api
from .dataset_source import DatasetSource
from .types import FieldSchemaDict, JSONDict


def upload_dataset(
    api_key: str,
    dataset_source: DatasetSource,
    *,
    schema_overrides: Optional[FieldSchemaDict] = None,
    modality: Optional[str] = None,
    id_column: Optional[str] = None,
) -> str:
    upload_id = upload_dataset_file(api_key, dataset_source)
    schema = get_proposed_schema(api_key, upload_id)

    if (schema is None or schema.get("immutable", False)) and (
        schema_overrides is not None or modality is not None or id_column is not None
    ):
        raise ValueError(
            "Schema_overrides, modality, and id_column parameters cannot be provided for simple zip uploads"
        )

    if schema is not None and not schema.get("immutable", False):
        schema["metadata"]["name"] = dataset_source.dataset_name
        if schema_overrides is not None:
            for field in schema_overrides:
                schema["fields"][field] = schema_overrides[field]
        if modality is not None:
            schema["metadata"]["modality"] = modality
        if id_column is not None:
            schema["metadata"]["id_column"] = id_column

    api.confirm_schema(api_key, schema, upload_id)
    dataset_id = get_ingestion_result(api_key, upload_id)
    return dataset_id


async def _upload_file_chunk_async(
    session: aiohttp.ClientSession,
    chunk: bytes,
    presigned_post: str,
) -> CIMultiDictProxy[str]:
    async with session.put(presigned_post, data=chunk) as response:
        return response.headers


async def upload_file_parts_async(
    dataset_source: DatasetSource, part_sizes: List[int], presigned_posts: List[str]
) -> List[JSONDict]:
    tasks = []
    chunks = dataset_source.get_chunks(part_sizes)
    async with aiohttp.ClientSession() as session:
        for chunk, presigned_post in zip(chunks, presigned_posts):
            tasks.append(
                asyncio.create_task(_upload_file_chunk_async(session, chunk, presigned_post))
            )
        task_results = await asyncio.gather(*tasks)
    return [
        {"ETag": json.loads(res.get("etag")), "PartNumber": i + 1}
        for i, res in enumerate(task_results)
    ]


def upload_file_parts(
    dataset_source: DatasetSource, part_sizes: List[int], presigned_posts: List[str]
) -> List[JSONDict]:
    responses = [
        requests.put(presigned_post, data=chunk)
        for chunk, presigned_post in tqdm(
            list(zip(dataset_source.get_chunks(part_sizes), presigned_posts)),
            desc="Uploading dataset...",
            bar_format="{desc}: {percentage:3.0f}%|{bar}|",
        )
    ]
    return [
        {"ETag": json.loads(res.headers["etag"]), "PartNumber": i + 1}
        for i, res in enumerate(responses)
    ]


def upload_dataset_file(api_key: str, dataset_source: DatasetSource) -> str:
    upload_id, part_sizes, presigned_posts = api.initialize_upload(
        api_key,
        dataset_source.get_filename(),
        dataset_source.file_type,
        dataset_source.file_size,
    )
    upload_parts = upload_file_parts(dataset_source, part_sizes, presigned_posts)
    api.complete_file_upload(api_key, upload_id, upload_parts)
    return upload_id


def get_proposed_schema(api_key: str, upload_id: str) -> Optional[JSONDict]:
    res = api.poll_progress(
        upload_id,
        functools.partial(api.get_proposed_schema, api_key),
        "Generating schema...",
    )
    schema = res.get("schema")
    return schema


def get_ingestion_result(
    api_key: str,
    upload_id: str,
) -> str:
    api.poll_progress(
        upload_id,
        functools.partial(api.get_ingestion_status, api_key),
        "Ingesting Dataset...",
    )
    res = api.get_dataset_id(api_key, upload_id)
    return str(res["dataset_id"])



"""
Contains utility functions for interacting with dataset files
"""
import io
import json
import os
import pathlib
import pandas as pd
from typing import Dict, List, Generator, IO, Union

from cleanlab_studio.cli.classes import CsvDataset, JsonDataset, ExcelDataset
from cleanlab_studio.cli.classes.dataset import Dataset
from cleanlab_studio.cli.types import (
    RecordType,
    DatasetFileExtension,
    ImageFileExtension,
)


def get_dataset_file_extension(filename: str) -> DatasetFileExtension:
    file_extension = pathlib.Path(filename).suffix.lower()
    return DatasetFileExtension(file_extension)


def _standardize_image_file_extension(file_extension: str) -> str:
    file_extension = file_extension.lower()
    if file_extension == ".jpg":
        return ".jpeg"
    return file_extension


def get_image_file_extension(filename: str) -> ImageFileExtension:
    file_extension = pathlib.Path(filename).suffix
    return ImageFileExtension(_standardize_image_file_extension(file_extension))


def get_filename(filepath: str) -> str:
    return os.path.split(filepath)[-1]


def get_file_size(filepath: str, ignore_missing_files: bool = False) -> int:
    if ignore_missing_files:
        try:
            return os.path.getsize(filepath)
        except IOError:
            return 0
    return os.path.getsize(filepath)


def read_file_as_df(filepath: str) -> pd.DataFrame:
    ext = get_dataset_file_extension(filepath)
    if ext == DatasetFileExtension.json:
        df = pd.read_json(filepath, convert_axes=False, convert_dates=False).T
        df.index = df.index.astype("str")
        df["id"] = df.index
    elif ext == DatasetFileExtension.csv:
        df = pd.read_csv(filepath, keep_default_na=True)
    else:  # ext == DatasetFileExtension.xls or ext == DatasetFileExtension.xlsx:
        df = pd.read_excel(filepath, keep_default_na=True)
    # elif ext in [".parquet"]:
    #     df = pd.read_parquet(filepath)
    return df


def is_null_value(val: str) -> bool:
    return val is None or val == "" or pd.isna(val)


def init_dataset_from_filepath(filepath: str) -> Union[Dataset[IO[str]], Dataset[IO[bytes]]]:
    ext = get_dataset_file_extension(filepath)
    if ext == DatasetFileExtension.csv:
        return CsvDataset(filepath)
    elif ext == DatasetFileExtension.xls or ext == DatasetFileExtension.xlsx:
        return ExcelDataset(filepath, file_type=pathlib.Path(filepath).suffix[1:])
    elif ext == DatasetFileExtension.json:
        return JsonDataset(filepath)

    raise ValueError(f"filepath {filepath} does not have supported extension.")


def init_dataset_from_fileobj(
    fileobj: Union[IO[str], IO[bytes]], ext: DatasetFileExtension
) -> Union[Dataset[IO[str]], Dataset[IO[bytes]]]:
    """Initializes dataset from file object."""
    if ext == DatasetFileExtension.csv:
        assert isinstance(fileobj, io.TextIOBase)
        return CsvDataset(fileobj=fileobj)
    elif ext in [DatasetFileExtension.xls, DatasetFileExtension.xlsx]:
        assert isinstance(fileobj, io.BufferedIOBase)
        return ExcelDataset(fileobj=fileobj, file_type=ext.value[1:])
    elif ext == DatasetFileExtension.json:
        assert isinstance(fileobj, io.BufferedIOBase)
        return JsonDataset(fileobj=fileobj)

    raise ValueError(f"Extension {ext.value} is not supported.")


def dump_json(filepath: str, obj: object) -> None:
    with open(filepath, "w") as f:
        f.write(json.dumps(obj, indent=2))


def append_rows(rows: List[RecordType], filename: str) -> None:
    df = pd.DataFrame(rows)
    ext = get_dataset_file_extension(filename)
    if ext == DatasetFileExtension.csv:
        if not os.path.exists(filename):
            df.to_csv(filename, mode="w", index=False, header=True)
        else:
            df.to_csv(filename, mode="a", index=False, header=False)
    elif ext == DatasetFileExtension.xls or ext == DatasetFileExtension.xlsx:
        if not os.path.exists(filename):
            with pd.ExcelWriter(filename) as writer:
                df.to_excel(writer, index=False, header=True)
        else:
            with pd.ExcelWriter(filename, mode="a") as writer:
                df.to_excel(writer, index=False, header=False)


def get_dataset_chunks(
    filepath: str,
    id_column: str,
    ids_to_fields_to_values: Dict[str, RecordType],
    num_rows_per_chunk: int,
) -> Generator[List[RecordType], None, None]:
    dataset = init_dataset_from_filepath(filepath)
    chunk = []
    for r in dataset.read_streaming_records():
        row_id = str(r.get(id_column))
        if row_id:
            updates = ids_to_fields_to_values[row_id]
            r.update(updates)
        chunk.append(r)

        if len(chunk) >= num_rows_per_chunk:
            yield chunk
            chunk = []

    if chunk:
        yield chunk



import click
import json
import os
from cleanlab_studio.cli.types import CommandState
from cleanlab_studio.internal.settings import CleanlabSettings
from typing import Dict, Any, Optional

FILENAME = "state.json"


class PreviousState:
    def __init__(self) -> None:
        self.state: CommandState = dict(command=None, args=dict())
        self.load_state()

    def load_state(self) -> None:
        fp = self.get_filepath()
        if os.path.exists(fp):
            with open(fp, "r") as f:
                self.state = json.load(f)

    @staticmethod
    def get_filepath() -> str:
        return os.path.join(CleanlabSettings.get_cleanlab_dir(), FILENAME)

    def save_state(self) -> None:
        os.makedirs(os.path.dirname(self.get_filepath()), exist_ok=True)
        with open(self.get_filepath(), "w") as f:
            json.dump(self.state, f)

    def init_state(self, updates: CommandState) -> None:
        self.state = dict(command=None, args=dict())
        self.state["command"] = updates["command"]
        self.state["args"] = updates["args"]
        self.save_state()

    def update_args(self, args_dict: Dict[str, Any]) -> None:
        self.state["args"].update(args_dict)
        self.save_state()

    def get_state(self) -> CommandState:
        return self.state

    def same_command(self, command_name: str, args: Dict[str, Any]) -> bool:
        if self.get("command") == command_name:
            same = all(self.get_arg(k) == args.get(k, None) for k in args)
            return same
        return False

    def get(self, item: str) -> Optional[Any]:
        return self.state.get(item, None)

    def get_arg(self, item: str) -> Optional[Any]:
        return self.state["args"].get(item, None)


previous_state = click.make_pass_decorator(PreviousState, ensure=True)



import itertools
import time
from typing import Optional

from tqdm import tqdm

from cleanlab_studio.errors import CleansetError
from cleanlab_studio.internal.api import api


def poll_cleanset_status(api_key: str, cleanset_id: str, timeout: Optional[float] = None) -> None:
    start_time = time.time()
    res = api.get_cleanset_status(api_key, cleanset_id)
    spinner = itertools.cycle("|/-\\")

    with tqdm(
        total=res["total_steps"],
        desc="Cleanset Progress: \\",
        bar_format="{desc} Step {n_fmt}/{total_fmt}{postfix}",
    ) as pbar:
        while not res["is_ready"] and not res["has_error"]:
            if pbar.total is None and res["total_steps"] is not None:
                pbar.total = res["total_steps"]
                pbar.refresh()

            pbar.set_postfix_str(res["step_description"])
            pbar.update(int(res["step"]) - pbar.n)

            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError("Cleanset not ready before timeout")

            for _ in range(50):
                time.sleep(0.1)
                pbar.set_description_str(f"Cleanset Progress: {next(spinner)}")

            res = api.get_cleanset_status(api_key, cleanset_id)

        if res["is_ready"]:
            pbar.update(int(res["step"]) - pbar.n)
            pbar.set_postfix_str(res["step_description"])
            return

        if res["has_error"]:
            pbar.set_postfix_str(res["step_description"])
            raise CleansetError(f"Cleanset {cleanset_id} failed to complete")



import json
import pathlib
from typing import Optional
import os

from cleanlab_studio.internal.upload_helpers import (
    get_ingestion_result,
    get_proposed_schema,
    upload_dataset_file,
)

import click

from cleanlab_studio.cli import click_helpers
from cleanlab_studio.cli.click_helpers import abort, info, log
from cleanlab_studio.cli.dataset.schema_helpers import (
    load_schema,
    save_schema,
)
from cleanlab_studio.cli.decorators import auth_config
from cleanlab_studio.cli.decorators.auth_config import AuthConfig
from cleanlab_studio.internal import api
from cleanlab_studio.internal.dataset_source import FilepathDatasetSource
from cleanlab_studio.internal.types import JSONDict


@click.command(help="upload your dataset to Cleanlab Studio")
@click.argument(
    "filepath",
    type=click.Path(),
    required=False,
    default=None,
)
@click.option(
    "--schema_path",
    "-s",
    type=click.Path(),
    help="If uploading with a schema, specify the JSON schema filepath.",
)
@auth_config
def upload(
    config: AuthConfig,
    filepath: Optional[str],
    schema_path: Optional[str],
) -> None:
    api_key = config.get_api_key()

    if filepath is None:
        filepath = click_helpers.prompt_for_filepath("Specify your dataset filepath")
    if not os.path.exists(filepath):
        abort(f"cannot upload '{filepath}': no such file or directory")

    dataset_source = FilepathDatasetSource(filepath=pathlib.Path(filepath))
    upload_id = upload_dataset_file(api_key, dataset_source)

    schema: Optional[JSONDict]

    # Check if uploading with schema
    if schema_path is not None:
        schema = load_schema(schema_path)
        proceed_upload = True

    ### Propose schema
    else:
        schema = get_proposed_schema(api_key, upload_id)
        proceed_upload = None
        if schema is None or schema.get("immutable", False):
            proceed_upload = True
        else:
            log(json.dumps(schema, indent=2))
            info(f"No schema was provided. We propose the above schema based on your dataset.")

            proceed_upload = click.confirm("\nUse this schema?", default=True)
            if not proceed_upload:
                info(
                    "Proposed schema rejected. Please submit your own schema using --schema.\n",
                )

            save_filepath = click_helpers.confirm_save_prompt_filepath(
                save_message="Save the generated schema?",
                save_default=None,
                prompt_message="Specify a filename for the schema. Leave this blank to use default",
                prompt_default="schema.json",
                no_save_message="Schema was not saved.",
            )
            if save_filepath:
                save_schema(schema, save_filepath)

    if proceed_upload or schema is None:
        api.confirm_schema(api_key, schema, upload_id)
        dataset_id = get_ingestion_result(api_key, upload_id)
        log(f"Successfully uploaded dataset with ID: {dataset_id}")
        click_helpers.success(
            "Upload completed. View your uploaded dataset at https://app.cleanlab.ai"
        )



import jsonstreams
from typing import Dict

from cleanlab_studio.cli.types import RecordType, DatasetFileExtension
from cleanlab_studio.cli.util import get_dataset_file_extension, get_dataset_chunks, append_rows


def combine_fields_with_dataset(
    dataset_filepath: str,
    id_column: str,
    ids_to_fields_to_values: Dict[str, RecordType],
    output_filepath: str,
    num_rows_per_chunk: int = 10000,
) -> None:
    output_extension = get_dataset_file_extension(output_filepath)
    if output_extension == DatasetFileExtension.json:
        with jsonstreams.Stream(
            jsonstreams.Type.OBJECT, filename=output_filepath, indent=True, pretty=True
        ) as s:
            with s.subarray("rows") as rows:
                for chunk in get_dataset_chunks(
                    dataset_filepath, id_column, ids_to_fields_to_values, num_rows_per_chunk
                ):
                    for row in chunk:
                        rows.write(row)
    elif output_extension in [
        DatasetFileExtension.csv,
        DatasetFileExtension.xls,
        DatasetFileExtension.xlsx,
    ]:
        for chunk in get_dataset_chunks(
            dataset_filepath, id_column, ids_to_fields_to_values, num_rows_per_chunk
        ):
            append_rows(chunk, output_filepath)
    else:
        raise ValueError(f"Invalid file type: {output_extension}.")



from typing import cast, Literal, TypedDict
from cleanlab_studio.internal.api import api


valid_quality_presets = ["best", "high", "medium", "low", "base"]
QualityPreset = Literal["best", "high", "medium", "low", "base"]


class TLMResponse(TypedDict):
    response: str
    confidence_score: float


class TLM:
    """TLM interface class."""

    def __init__(self, api_key: str, quality_preset: QualityPreset) -> None:
        """Initializes TLM interface."""
        self._api_key = api_key

        if quality_preset not in valid_quality_presets:
            raise ValueError(
                f"Invalid quality preset {quality_preset} -- must be one of {valid_quality_presets}"
            )

        self._quality_preset = quality_preset

    def prompt(self, prompt: str) -> TLMResponse:
        """
        Get response and confidence from TLM.

        Args:
            prompt: prompt for the TLM
        Returns:
            A dict containing the TLM response
        """
        tlm_response = api.tlm_prompt(self._api_key, prompt, self._quality_preset)
        return {
            "response": tlm_response["response"],
            "confidence_score": tlm_response["confidence_score"],
        }

    def get_confidence_score(self, prompt: str, response: str) -> float:
        """Gets confidence score for prompt-response pair.

        Args:
            prompt: prompt for the TLM
            response: response for the TLM  to evaluate
        Returns:
            float corresponding to the TLM's confidence score
        """
        if self._quality_preset == "base":
            raise ValueError(
                "Cannot get confidence score with `base` quality_preset -- choose a higher preset."
            )

        return cast(
            float,
            api.tlm_get_confidence_score(self._api_key, prompt, response, self._quality_preset)[
                "confidence_score"
            ],
        )



import pathlib
from typing import List

import pandas as pd
import pytest

from cleanlab_studio.cli.classes.json_dataset import JsonDataset
from .constants import DATASETS_DIR
from .utils import assert_dataset_matches_dataframe, json_filepath_to_dataframe


json_DATASETS: List[pathlib.Path] = [*DATASETS_DIR.glob(r"*.json")]


@pytest.mark.parametrize(
    ("json_filepath", "json_df"),
    ((json_filepath, json_filepath_to_dataframe(json_filepath)) for json_filepath in json_DATASETS),
)
def test_json_dataset_from_filepath_matches_expected(
    json_filepath: pathlib.Path, json_df: pd.DataFrame
):
    """Tests that JsonDataset, loaded from filepath, matches json loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load jsonDataset from filepath
    json_dataset = JsonDataset(filepath=str(json_filepath))

    # check that methods match values read by pandas
    assert_dataset_matches_dataframe(json_dataset, json_df, ignore_id=True)


@pytest.mark.parametrize(
    ("json_filepath", "json_df"),
    ((json_filepath, json_filepath_to_dataframe(json_filepath)) for json_filepath in json_DATASETS),
)
def test_json_dataset_from_fileobj_matches_expected(
    json_filepath: pathlib.Path, json_df: pd.DataFrame
):
    """Tests that JsonDataset, loaded from file object, matches json loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load jsonDataset from object
    with open(json_filepath, **JsonDataset.READ_ARGS) as json_file:
        json_dataset = JsonDataset(fileobj=json_file)

        # check that methods match values read by pandas
        assert_dataset_matches_dataframe(json_dataset, json_df, ignore_id=True)



"""
Python API for Cleanlab Studio.
"""
from typing import Any, List, Literal, Optional, Union
import warnings

import numpy as np
import numpy.typing as npt
import pandas as pd

from . import inference
from . import trustworthy_llm
from cleanlab_studio.errors import CleansetError
from cleanlab_studio.internal import clean_helpers, upload_helpers
from cleanlab_studio.internal.api import api
from cleanlab_studio.internal.util import (
    init_dataset_source,
    check_none,
    check_not_none,
)
from cleanlab_studio.internal.settings import CleanlabSettings
from cleanlab_studio.internal.types import FieldSchemaDict

_pyspark_exists = api.pyspark_exists
if _pyspark_exists:
    import pyspark.sql


class Studio:
    """Used to interact with Cleanlab Studio."""

    _api_key: str

    def __init__(self, api_key: Optional[str]):
        if not api.is_valid_client_version():
            raise ValueError(
                "CLI is out of date and must be updated. Run 'pip install --upgrade cleanlab-studio'."
            )
        if api_key is None:
            try:
                api_key = CleanlabSettings.load().api_key
                if api_key is None:
                    raise ValueError
            except (FileNotFoundError, KeyError, ValueError):
                raise ValueError(
                    "No API key found; either specify API key or log in with 'cleanlab login' first"
                )
        api.validate_api_key(api_key)
        self._api_key = api_key

    def upload_dataset(
        self,
        dataset: Any,
        dataset_name: Optional[str] = None,
        *,
        schema_overrides: Optional[FieldSchemaDict] = None,
        modality: Optional[str] = None,
        id_column: Optional[str] = None,
    ) -> str:
        """
        Uploads a dataset to Cleanlab Studio.

        Args:
            dataset: Object representing the dataset to upload. Currently supported formats include a `str` path to your dataset, a pandas DataFrame, a pyspark DataFrame.
            dataset_name: Name for your dataset in Cleanlab Studio (optional if uploading from filepath).
            schema_overrides: Optional dictionary of overrides you would like to make to the schema of your dataset. If not provided, schema will be inferred. Format defined here: https://cleanlab-studio.readthedocs.io/en/latest/concepts/datasets.html#schemas
            modality: Optional parameter to override the modality of your dataset. If not provided, modality will be inferred.
            id_column: Optional parameter to override the ID column of your dataset. If not provided, a monotonically increasing ID column will be generated.

        Returns:
            ID of uploaded dataset.
        """
        ds = init_dataset_source(dataset, dataset_name)
        return upload_helpers.upload_dataset(
            self._api_key,
            ds,
            schema_overrides=schema_overrides,
            modality=modality,
            id_column=id_column,
        )

    def download_cleanlab_columns(
        self,
        cleanset_id: str,
        include_cleanlab_columns: bool = True,
        include_project_details: bool = False,
        to_spark: bool = False,
    ) -> Any:
        """
        Downloads [Cleanlab columns](/guide/concepts/cleanlab_columns/) for a cleanset.

        Args:
            cleanset_id: ID of cleanset to download columns from. To obtain cleanset ID from project ID use, [get_latest_cleanset_id](#method-get_latest_cleanset_id).
            include_action: Whether to include a column with any actions taken on the cleanset in the downloaded columns.

        Returns:
            A pandas or pyspark DataFrame. Type is `Any` to avoid requiring pyspark installation.
        """
        rows_df = api.download_cleanlab_columns(
            self._api_key,
            cleanset_id,
            include_cleanlab_columns=include_cleanlab_columns,
            include_project_details=include_project_details,
            to_spark=to_spark,
        )
        if "cleanlab_row_ID" in rows_df.columns:
            rows_df.sort_values(by="cleanlab_row_ID")
        return rows_df

    def apply_corrections(self, cleanset_id: str, dataset: Any, keep_excluded: bool = False) -> Any:
        """
        Applies corrections from a Cleanlab Studio cleanset to your dataset. Corrections can be made by viewing your project in the Cleanlab Studio webapp (see [Cleanlab Studio web quickstart](/guide/quickstart/web#review-issues-detected-in-your-dataset-and-correct-them)).

        Args:
            cleanset_id: ID of cleanset to apply corrections from.
            dataset: Dataset to apply corrections to. Supported formats include pandas DataFrame and pyspark DataFrame. Dataset should have the same number of rows as the dataset used to create the project. It should also contain a label column with the same name as the label column for the project.
            keep_excluded: Whether to retain rows with an "exclude" action. By default these rows will be removed from the dataset.

        Returns:
            A copy of the dataset with corrections applied.
        """
        project_id = api.get_project_of_cleanset(self._api_key, cleanset_id)
        label_column = api.get_label_column_of_project(self._api_key, project_id)
        id_col = api.get_id_column(self._api_key, cleanset_id)
        if _pyspark_exists and isinstance(dataset, pyspark.sql.DataFrame):
            from pyspark.sql.functions import udf

            cl_cols = self.download_cleanlab_columns(cleanset_id, to_spark=True)
            corrected_ds_spark = dataset.alias("corrected_ds")
            if id_col not in corrected_ds_spark.columns:
                from pyspark.sql.functions import (
                    row_number,
                    monotonically_increasing_id,
                )
                from pyspark.sql.window import Window

                corrected_ds_spark = corrected_ds_spark.withColumn(
                    id_col,
                    row_number().over(Window.orderBy(monotonically_increasing_id())) - 1,
                )
            both = cl_cols.select([id_col, "action", "clean_label"]).join(
                corrected_ds_spark.select([id_col, label_column]),
                on=id_col,
                how="left",
            )
            final = both.withColumn(
                "__cleanlab_final_label",
                # XXX hacky, checks if label is none by hand
                # instead, use original JSON, which uses null values where it's not specified
                udf(lambda original, clean: original if check_none(clean) else clean)(
                    both[label_column],
                    "clean_label",
                ),
            )
            new_labels = final.select(
                [id_col, "action", "__cleanlab_final_label"]
            ).withColumnRenamed("__cleanlab_final_label", label_column)
            return (
                corrected_ds_spark.drop(label_column)
                .join(new_labels, on=id_col, how="right")
                .where(new_labels["action"] != "exclude")
                .drop("action")
            )
        elif isinstance(dataset, pd.DataFrame):
            cl_cols = self.download_cleanlab_columns(cleanset_id)
            joined_ds: pd.DataFrame
            if id_col in dataset.columns:
                joined_ds = dataset.join(cl_cols.set_index(id_col), on=id_col)
            else:
                joined_ds = dataset.join(cl_cols.set_index(id_col).sort_values(by=id_col))
            joined_ds["__cleanlab_final_label"] = joined_ds["clean_label"].where(
                np.asarray(list(map(check_not_none, joined_ds["clean_label"].to_numpy()))),
                dataset[label_column].to_numpy(),
            )

            corrected_ds: pd.DataFrame = dataset.copy()
            corrected_ds[label_column] = joined_ds["__cleanlab_final_label"]
            if not keep_excluded:
                corrected_ds = corrected_ds.loc[(joined_ds["action"] != "exclude").fillna(True)]
            else:
                corrected_ds["action"] = joined_ds["action"]
            return corrected_ds

        else:
            raise ValueError(
                f"Provided unsupported dataset of type: {type(dataset)}. We currently support applying corrections to pandas or pyspark dataframes"
            )

    def create_project(
        self,
        dataset_id: str,
        project_name: str,
        modality: Literal["text", "tabular", "image"],
        *,
        task_type: Literal["multi-class", "multi-label"] = "multi-class",
        model_type: Literal["fast", "regular"] = "regular",
        label_column: Optional[str] = None,
        feature_columns: Optional[List[str]] = None,
        text_column: Optional[str] = None,
    ) -> str:
        """
        Creates a Cleanlab Studio project.

        Args:
            dataset_id: ID of dataset to create project for.
            project_name: Name for resulting project.
            modality: Modality of project (i.e. text, tabular, image).
            task_type: Type of classification to perform (i.e. multi-class, multi-label).
            model_type: Type of model to train (i.e. fast, regular).
            label_column: Name of column in dataset containing labels (if not supplied, we'll make our best guess).
            feature_columns: List of columns to use as features when training tabular modality project (if not supplied and modality is "tabular" we'll use all valid feature columns).
            text_column: Name of column containing the text to train text modality project on (if not supplied and modality is "text" we'll make our best guess).

        Returns:
            ID of created project.
        """
        dataset_details = api.get_dataset_details(self._api_key, dataset_id)

        if label_column is not None:
            if label_column not in dataset_details["label_columns"]:
                raise ValueError(
                    f"Invalid label column: {label_column}. Label column must have categorical feature type"
                )
        else:
            label_column = str(dataset_details["label_column_guess"])
            print(f"Label column not supplied. Using best guess {label_column}")

        if feature_columns is not None and modality != "tabular":
            if label_column in feature_columns:
                raise ValueError("Label column cannot be included in feature columns")
            raise ValueError("Feature columns supplied, but project modality is not tabular")
        if feature_columns is None:
            if modality == "tabular":
                feature_columns = dataset_details["distinct_columns"]
                feature_columns.remove(label_column)
                print(f"Feature columns not supplied. Using all valid feature columns")

        if text_column is not None:
            if modality != "text":
                raise ValueError("Text column supplied, but project modality is not text")
            elif text_column not in dataset_details["text_columns"]:
                raise ValueError(
                    f"Invalid text column: {text_column}. Column must have text feature type"
                )
        if text_column is None and modality == "text":
            text_column = dataset_details["text_column_guess"]
            print(f"Text column not supplied. Using best guess {text_column}")

        return api.clean_dataset(
            api_key=self._api_key,
            dataset_id=dataset_id,
            project_name=project_name,
            task_type=task_type,
            modality=modality,
            model_type=model_type,
            label_column=label_column,
            feature_columns=feature_columns if feature_columns is not None else [],
            text_column=text_column,
        )

    def poll_cleanset_status(self, cleanset_id: str, timeout: Optional[int] = None) -> bool:
        """
        Repeatedly polls for cleanset status while the cleanset is being generated. Blocks until cleanset is ready, there is a cleanset error, or `timeout` is exceeded.

        Args:
            cleanset_id: ID of cleanset to check status of.
            timeout: Optional timeout after which to stop polling for progress. If not provided, will block until cleanset is ready.

        Returns:
            After cleanset is done being generated, returns `True` if cleanset is ready to use, `False` otherwise.
        """
        warnings.warn(
            "Poll cleanset status method has been deprecated -- please use wait_for_cleanset_ready method instead.",
            DeprecationWarning,
        )

        try:
            clean_helpers.poll_cleanset_status(self._api_key, cleanset_id, timeout)
            return True

        except (TimeoutError, CleansetError):
            return False

    def wait_until_cleanset_ready(self, cleanset_id: str, timeout: Optional[float] = None) -> None:
        """Blocks until a cleanset is ready or the timeout is reached.

        Args:
            cleanset_id (str): ID of cleanset to check status for.
            timeout (Optional[float], optional): timeout for polling, in seconds. Defaults to None.

        Raises:
            TimeoutError: if cleanset is not ready by end of timeout
            CleansetError: if cleanset errored while running
        """
        clean_helpers.poll_cleanset_status(self._api_key, cleanset_id, timeout)

    def get_latest_cleanset_id(self, project_id: str) -> str:
        """
        Gets latest cleanset ID for a project.

        Args:
            project_id: ID of project.

        Returns:
            ID of latest associated cleanset.
        """
        return api.get_latest_cleanset_id(self._api_key, project_id)

    def delete_project(self, project_id: str) -> None:
        """
        Deletes a project from Cleanlab Studio.

        Args:
            project_id: ID of project to delete.
        """
        api.delete_project(self._api_key, project_id)
        print(f"Successfully deleted project: {project_id}")

    def get_model(self, model_id: str) -> inference.Model:
        """
        Gets a model deployed by Cleanlab Studio.

        Args:
            model_id: ID of model to get. This ID should be fetched in the deployments page of the app UI.

        Returns:
            [Model](../inference#class-model) object with methods to run predictions on new input data.
        """
        return inference.Model(self._api_key, model_id)

    def download_pred_probs(
        self,
        cleanset_id: str,
        keep_id: bool = False,
    ) -> pd.DataFrame:
        """
        Downloads predicted probabilities for a cleanset

        The probabilities will be returned as a `pd.DataFrame`. If `keep_id` is `True`,
        the DataFrame will include an ID column that can be used for database joins/merges with
        the original dataset or downloaded Cleanlab columns.
        """
        pred_probs: Union[npt.NDArray[np.float_], pd.DataFrame] = api.download_array(
            self._api_key, cleanset_id, "pred_probs"
        )
        if not isinstance(pred_probs, pd.DataFrame):
            pred_probs = pd.DataFrame(pred_probs)
            return pred_probs

        if not keep_id:
            id_col = api.get_id_column(self._api_key, cleanset_id)
            if id_col in pred_probs.columns:
                pred_probs = pred_probs.drop(id_col, axis=1)

        return pred_probs

    def download_embeddings(
        self,
        cleanset_id: str,
    ) -> npt.NDArray[np.float_]:
        """
        Downloads embeddings for a cleanset
        """
        return np.asarray(api.download_array(self._api_key, cleanset_id, "embeddings"))

    def TLM(self, *, quality_preset: trustworthy_llm.QualityPreset = "low") -> trustworthy_llm.TLM:
        """Gets Trustworthy Language Model (TLM) object to prompt.

        Returns:
            TLM: the Trustworthy Language Model object
        """
        return trustworthy_llm.TLM(self._api_key, quality_preset)



import pathlib

from PIL import Image


def get_image_filepath(base_directory: pathlib.Path, image_filepath: pathlib.Path) -> pathlib.Path:
    """Joins base directory with relative or absolute image filepath.

    Note
    ----
    if column_value (image filepath) is an absolute path, base directory will be ignored
    """
    return base_directory.joinpath(image_filepath)


def image_file_exists(image_filepath: pathlib.Path) -> bool:
    return image_filepath.exists()


def image_file_readable(image_filepath: pathlib.Path) -> bool:
    """
    readable == image file can be opened by Pillow
    """
    try:
        Image.open(image_filepath)
        return True

    except IOError:
        return False



from decimal import Decimal
import pathlib
from typing import Any

import numpy as np
import pandas as pd

from cleanlab_studio.cli.classes.csv_dataset import CsvDataset
from cleanlab_studio.cli.classes.dataset import Dataset
from cleanlab_studio.cli.classes.json_dataset import JsonDataset
from cleanlab_studio.cli.classes.excel_dataset import ExcelDataset


def csv_filepath_to_dataframe(csv_filepath: pathlib.Path) -> pd.DataFrame:
    """Reads CSV from filepath to dataframe."""
    with open(csv_filepath, **CsvDataset.READ_ARGS) as csv_file:
        return pd.read_csv(csv_file, keep_default_na=True)


def excel_filepath_to_dataframe(excel_filepath: pathlib.Path) -> pd.DataFrame:
    """Reads excel from filepath to dataframe."""
    with open(excel_filepath, **ExcelDataset.READ_ARGS) as excel_file:
        return pd.read_excel(excel_file, keep_default_na=True)


def json_filepath_to_dataframe(json_filepath: pathlib.Path) -> pd.DataFrame:
    """Reads json from filepath to dataframe."""
    with open(json_filepath, **JsonDataset.READ_ARGS) as json_file:
        return pd.read_json(json_file, orient="records", convert_axes=False, convert_dates=False)


def assert_dataset_matches_dataframe(
    dataset: Dataset, dataframe: pd.DataFrame, ignore_id: bool = False
) -> None:
    """Asserts that Cleanlab Studio Dataset object matches dataframe loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    assert len(dataset) == dataframe.shape[0], "Dataset length does not match dataframe length"

    assert np.all(
        dataset.get_columns() == dataframe.columns.values
    ), "Dataset columns do not match dataframe columns"

    for dataset_row, (_, dataframe_row) in zip(
        dataset.read_streaming_records(), dataframe.iterrows()
    ):
        for dataset_val, dataframe_val in zip(dataset_row.values(), dataframe_row.values):
            assert_dataset_val_matches_dataframe_val(
                dataset_val,
                dataframe_val,
                assert_msg="Dataset read_streaming_records does not match dataframe values.",
            )

    for dataset_row, (_, dataframe_row) in zip(
        dataset.read_streaming_values(), dataframe.iterrows()
    ):
        for dataset_val, dataframe_val in zip(dataset_row, dataframe_row.values):
            assert_dataset_val_matches_dataframe_val(
                dataset_val,
                dataframe_val,
                assert_msg="Dataset read_streaming_values does not match dataframe values.",
            )

    dataset_dataframe = dataset.read_file_as_dataframe()
    if ignore_id:
        dataset_dataframe = dataset_dataframe.drop(columns=["id"])

    assert dataset_dataframe.reset_index(drop=True).equals(
        dataframe.reset_index(drop=True)
    ), "Dataset read as dataframe does not match dataframe"


def assert_dataset_val_matches_dataframe_val(
    dataset_val: Any, dataframe_val: Any, assert_msg: str
) -> None:
    """Asserts that dataset value matches dataframe value.

    Ignores nans.
    If dataset val is int and dataframe val is float, converts dataframe val to int.
    If dataset val is float and dataframe val is int, converts dataset val to int.
    If dataset val is Decimal and dataframe val is float, converts dataset val to float (and round both).
    """
    if pd.isnull(dataframe_val):
        return

    if isinstance(dataframe_val, float):
        if isinstance(dataset_val, int):
            dataframe_val = int(dataframe_val)
        elif isinstance(dataset_val, Decimal):
            dataset_val = round(float(dataset_val), ndigits=5)
            dataframe_val = round(dataframe_val, ndigits=5)

    elif isinstance(dataframe_val, int):
        if isinstance(dataset_val, float):
            dataset_val = int(dataset_val)

    assert str(dataset_val) == str(dataframe_val), assert_msg



import os
from collections import defaultdict
from typing import Dict, Any, Optional

import click
import pandas as pd

from cleanlab_studio.cli import api_service
from cleanlab_studio.cli import click_helpers
from cleanlab_studio.cli import util
from cleanlab_studio.cli.cleanset.download_helpers import combine_fields_with_dataset
from cleanlab_studio.cli.click_helpers import log, progress
from cleanlab_studio.cli.decorators import previous_state, auth_config
from cleanlab_studio.cli.decorators.auth_config import AuthConfig
from cleanlab_studio.cli.decorators.previous_state import PreviousState
from cleanlab_studio.cli.types import RecordType, DatasetFileExtension
from cleanlab_studio.internal.settings import CleanlabSettings
from cleanlab_studio.internal.api import api


@click.command(help="download Cleanlab columns")
@click.option(
    "--id",
    type=str,
    prompt=True,
    help="cleanset ID",
)
@click.option(
    "--filepath",
    "-f",
    type=click.Path(),
    help=(
        "To combine Cleanlab columns with existing dataset, set this as the filepath to the"
        " original dataset"
    ),
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    help="Output filepath",
)
@click.option(
    "--all",
    "-a",
    is_flag=True,
    help=("Set this flag to download all available data."),
)
@click.option(
    "--cleanlab_columns",
    "-c",
    is_flag=True,
    help=(
        "Set this flag to download all Cleanlab columns (suggested label, clean label, label"
        " quality, issue, outlier, ambiguous, near duplicate, well-labeled)."
    ),
)
@click.option(
    "--project_details",
    "-p",
    is_flag=True,
    help=(
        "Set this flag to download project information for a cleanset"
        " (action, previously_resolved, top_labels, top_probs, p_suggested_label,"
        "predicted_label, p_given_label, unlabeled (initially))"
    ),
)
@previous_state
@auth_config
def download(
    config: AuthConfig,
    prev_state: PreviousState,
    id: str,
    filepath: Optional[str],
    output: Optional[str],
    all: bool,
    cleanlab_columns: bool,
    project_details: bool,
) -> None:
    prev_state.init_state(dict(command="download labels", args=dict(id=id)))
    CleanlabSettings.init_cleanlab_dir()
    api_key = config.get_api_key()
    id_column = api_service.get_id_column(api_key, cleanset_id=id)

    progress("Downloading Cleanlab columns...")
    clean_df = api_service.download_cleanlab_columns(
        api_key,
        cleanset_id=id,
        include_cleanlab_columns=cleanlab_columns or all,
        include_project_details=project_details or all,
    )
    clean_df = clean_df.set_index(id_column)
    clean_df = drop_action_col(clean_df)

    if filepath:
        if not os.path.exists(filepath):
            log(f"Specified file {filepath} could not be found.")
            filepath = click_helpers.prompt_for_filepath("Specify your dataset filepath")

        filename = util.get_filename(filepath)
        while output is None:
            output = click.prompt(
                "Specify your output filepath. Leave blank to use default",
                default=f"cleaned_{filename}",
            )
            if os.path.exists(output):
                click_helpers.error(
                    "A file already exists at this filepath, use a different filepath."
                )
                output = None

        ids_to_fields_to_values: Dict[str, RecordType] = defaultdict(dict)
        for row_id, row in clean_df.iterrows():
            fields_to_values = dict(row)
            ids_to_fields_to_values[str(row_id)] = fields_to_values

        combine_fields_with_dataset(filepath, id_column, ids_to_fields_to_values, output)
        click_helpers.success(f"Saved to {output}")
    else:
        while output is None or util.get_dataset_file_extension(output) != DatasetFileExtension.csv:
            output = click.prompt(
                "Specify your output filepath (must be .csv). Leave blank to use default",
                default=f"clean_labels.csv",
            )
        clean_df.to_csv(output)
        click_helpers.success(f"Saved to {output}")


def drop_action_col(dataframe: pd.DataFrame) -> pd.DataFrame:
    if "action" in dataframe.columns:
        return dataframe.drop("action", axis=1)
    return dataframe



from typing import Optional, List

import click
from cleanlab_studio.cli.dataset.schema_helpers import _find_best_matching_column


def get_id_column_if_undefined(id_column: Optional[str], columns: List[str]) -> str:
    if id_column is None:
        id_column_guess = _find_best_matching_column("id", columns)
        while id_column not in columns:
            id_column = click.prompt(
                "Specify the name of the ID column in your dataset", default=id_column_guess
            )
    return id_column


def get_filepath_column_if_undefined(
    modality: str, columns: List[str], filepath_column: Optional[str]
) -> str:
    if filepath_column is None:
        filepath_column_guess = _find_best_matching_column("filepath", columns)
        while filepath_column not in columns:
            filepath_column = click.prompt(
                f"Specify the name of the filepath column (containing your {modality} filepaths) in your dataset",
                default=filepath_column_guess,
            )
    return filepath_column



import abc
import csv
import io
import time
from typing import List, Tuple, Union
from typing_extensions import TypeAlias

import numpy as np
import numpy.typing as npt
import pandas as pd

from cleanlab_studio.internal.api import api


TextBatch = Union[List[str], npt.NDArray[np.str_], pd.Series]
TabularBatch: TypeAlias = pd.DataFrame
Batch = Union[TextBatch, TabularBatch]

Predictions = Union[npt.NDArray[np.int_], npt.NDArray[np.str_]]
ClassProbablities: TypeAlias = pd.DataFrame


class Model(abc.ABC):
    """Base class for deployed model inference."""

    def __init__(self, api_key: str, model_id: str):
        """Initializes model class w/ API key and model ID."""
        self._api_key = api_key
        self._model_id = model_id

    def predict(
        self,
        batch: Batch,
        return_pred_proba: bool = False,
        timeout: int = 600,
    ) -> Union[Predictions, Tuple[Predictions, ClassProbablities]]:
        """
        Gets predictions for batch of examples.

        Args:
            batch: batch of examples to predict classes for
            return_pred_proba: if should return class probabilities for each example
            timeout: optional parameter to set timeout for predictions in seconds

        Returns:
            predictions from batch as a numpy array, optionally also pandas dataframe of class probabilties
        """
        csv_batch = self._convert_batch_to_csv(batch)
        predictions, class_probabilities = self._predict_from_csv(csv_batch, timeout)

        if return_pred_proba:
            return predictions, class_probabilities

        return predictions

    def _predict_from_csv(
        self, batch: io.StringIO, timeout: int
    ) -> Tuple[Predictions, ClassProbablities]:
        """Gets predictions for batch of examples.

        :param batch: batch of example to predict classes for, as in-memory CSV file
        :return: predictions from batch, class probabilities
        """
        query_id: str = api.upload_predict_batch(self._api_key, self._model_id, batch)
        api.start_prediction(self._api_key, self._model_id, query_id)

        # Set timeout to prevent users from getting stuck indefinitely when there is a failure
        timeout_limit = time.time() + timeout
        while time.time() < timeout_limit:
            resp = api.get_prediction_status(self._api_key, query_id)

            if result_url := resp.get("results"):
                results: pd.DataFrame = pd.read_csv(result_url)
                return results.pop("Suggested Label").to_numpy(), results

            time.sleep(1)

        else:
            raise TimeoutError(f"Timeout of {timeout}s expired while waiting for prediction")

    @staticmethod
    def _convert_batch_to_csv(batch: Batch) -> io.StringIO:
        """Converts batch object to CSV string IO."""
        sio = io.StringIO()

        # handle text batches
        if isinstance(batch, (list, np.ndarray, pd.Series)):
            writer = csv.writer(sio)

            # write header
            writer.writerow(["text"])

            # write labels to CSV
            for input_data in batch:
                writer.writerow([input_data])

        # handle tabular batches
        elif isinstance(batch, pd.DataFrame):
            batch.to_csv(sio)

        else:
            raise TypeError(f"Invalid type of batch: {type(batch)}")

        sio.seek(0)
        return sio



import click
from cleanlab_studio.cli.dataset.upload import upload


@click.group(help="upload datasets, generate & validate dataset schema")
def dataset() -> None:
    pass


dataset.add_command(upload)



import io
from typing import IO

import pandas as pd

try:
    import pyspark.sql
except ImportError:
    raise ImportError(
        'Must install pyspark to upload from pyspark dataframe. Use "pip install pyspark"'
    )

from .dataframe_dataset_source import DataFrameDatasetSource


class PySparkDatasetSource(DataFrameDatasetSource[pyspark.sql.DataFrame]):
    def _init_fileobj_from_df(self, df: pyspark.sql.DataFrame) -> IO[bytes]:
        fileobj = io.BytesIO()
        pd_df: pd.DataFrame = df.toPandas()
        pd_df.to_csv(fileobj, index=False)
        fileobj.seek(0)
        return fileobj



"""
Standardized methods for Click terminal logs

Error / failure: red
Warning: orange
Progress status: blue
Information: default
Success/completion: green
"""
import os

import click
from click import ClickException, secho, style, echo
from typing import Optional, Any


def abort(message: str) -> None:
    """
    Abort on error
    """
    raise ClickException(style(message, fg="red"))


def error(message: str) -> None:
    secho(style(message, fg="red"))


def success(message: str) -> None:
    secho(style(message, fg="green"))


def progress(message: str) -> None:
    secho(style(message, fg="yellow"))


def info(message: str) -> None:
    secho(style(message, fg="blue"))


def warn(message: str) -> None:
    secho(style(message, fg="magenta"))


def log(message: str) -> None:
    echo(message)


def prompt_for_filepath(prompt_message: str, default: Optional[Any] = None) -> str:
    filepath = str(click.prompt(prompt_message, default=default))
    while not os.path.exists(filepath):
        error(f"No file exists at: {filepath}")
        filepath = str(click.prompt(prompt_message))
    return filepath


def confirm_open_file(message: str, filepath: str) -> None:
    edit = click.confirm(message, default=None)
    if edit:
        click.edit(filename=filepath)


def confirm_save_data(message: str, default: Optional[Any] = None) -> bool:
    save = click.confirm(message, default=default)
    return save


def confirm_save_prompt_filepath(
    save_message: str,
    save_default: Optional[str],
    prompt_message: str,
    prompt_default: str,
    no_save_message: str,
) -> Optional[str]:
    save = confirm_save_data(save_message, default=save_default)
    if save:
        output: str = click.prompt(prompt_message, default=prompt_default)
        return output
    else:
        info(no_save_message)
        return None



from enum import Enum
from typing import Dict, Any, Optional, Union, TypedDict

JSONDict = Dict[str, Any]
IDType = Union[str, int]


class Modality(Enum):
    text = "text"
    tabular = "tabular"
    image = "image"


MODALITIES = [m.value for m in Modality]
MEDIA_MODALITIES = [Modality.image]


class DatasetFileExtension(Enum):
    csv = ".csv"
    xls = ".xls"
    xlsx = ".xlsx"
    json = ".json"


class ImageFileExtension(Enum):
    jpeg = ".jpeg"
    png = ".png"


RecordType = Dict[str, Any]


class CommandState(TypedDict):
    command: Optional[str]
    args: Dict[str, Optional[str]]


class CleanlabSettingsDict(TypedDict):
    version: Optional[str]
    api_key: Optional[str]



from enum import Enum
from typing import Dict, List, Collection
from dataclasses import dataclass


class ValidationWarning(Enum):
    MISSING_ID = "MISSING_ID"
    MISSING_VAL = "MISSING_VAL"
    MISSING_FILE = "MISSING_FILE"
    UNREADABLE_FILE = "UNREADABLE_FILE"
    TYPE_MISMATCH = "TYPE_MISMATCH"
    DUPLICATE_ID = "DUPLICATE_ID"


def warning_to_readable_name(warning: ValidationWarning) -> str:
    return {
        ValidationWarning.MISSING_ID: "Rows with missing IDs (rows are dropped)",
        ValidationWarning.MISSING_FILE: "Rows with non-existent filepaths (rows are dropped)",
        ValidationWarning.MISSING_VAL: "Rows with missing values (values replaced with null)",
        ValidationWarning.UNREADABLE_FILE: "Rows with unreadable files (rows are dropped)",
        ValidationWarning.TYPE_MISMATCH: (
            "Rows with values that do not match the schema (values replaced with null)"
        ),
        ValidationWarning.DUPLICATE_ID: (
            "Rows with duplicate IDs (only the first row instance is kept, all later rows dropped)"
        ),
    }[warning]


RowWarningsType = Dict[ValidationWarning, List[str]]


@dataclass
class WarningLog:
    MISSING_ID: List[str]
    MISSING_VAL: Dict[str, List[str]]
    MISSING_FILE: Dict[str, List[str]]
    UNREADABLE_FILE: Dict[str, List[str]]
    TYPE_MISMATCH: Dict[str, List[str]]
    DUPLICATE_ID: Dict[str, List[str]]

    @staticmethod
    def init() -> "WarningLog":
        return WarningLog(
            MISSING_ID=list(),
            MISSING_VAL=dict(),
            MISSING_FILE=dict(),
            UNREADABLE_FILE=dict(),
            TYPE_MISMATCH=dict(),
            DUPLICATE_ID=dict(),
        )

    def get(self, key: ValidationWarning) -> Collection[str]:
        return {
            ValidationWarning.MISSING_ID: self.MISSING_ID,
            ValidationWarning.MISSING_VAL: self.MISSING_VAL,
            ValidationWarning.MISSING_FILE: self.MISSING_FILE,
            ValidationWarning.UNREADABLE_FILE: self.UNREADABLE_FILE,
            ValidationWarning.TYPE_MISMATCH: self.TYPE_MISMATCH,
            ValidationWarning.DUPLICATE_ID: self.DUPLICATE_ID,
        }[key]

    def to_dict(self, readable: bool = False) -> Dict[str, Collection[str]]:
        retval = dict()
        for warning_type in ValidationWarning:
            key = warning_to_readable_name(warning_type) if readable else warning_type.value
            retval[key] = self.get(warning_type)
        return retval



from abc import abstractmethod
import contextlib
import pathlib
from typing import Optional, List, IO, Iterator


class DatasetSource:
    dataset_name: str
    file_size: int
    file_type: str
    _fileobj: Optional[IO[bytes]]
    _filepath: Optional[pathlib.Path]

    def __init__(self) -> None:
        self._filepath = None
        self._fileobj = None

    @contextlib.contextmanager
    def fileobj(self) -> Iterator[IO[bytes]]:
        """Yields open IO object to dataset file."""
        if self._filepath is not None:
            with open(self._filepath, "rb") as dataset_file:
                yield dataset_file

        elif self._fileobj is not None:
            try:
                yield self._fileobj
            finally:
                self._fileobj.seek(0)

        else:
            raise ValueError("Cannot return file object -- no filepath or fileobj available.")

    def get_chunks(self, chunk_sizes: List[int]) -> Iterator[bytes]:
        with self.fileobj() as f:
            for chunk_size in chunk_sizes:
                yield f.read(chunk_size)

    @abstractmethod
    def get_filename(self) -> str:
        pass



from abc import abstractmethod
import os
import pathlib
from typing import IO, Any, Generic, TypeVar, Union

import pandas as pd

try:
    import pyspark.sql

    pyspark_exists = True
except ImportError:
    pyspark_exists = False

from .dataset_source import DatasetSource

df_type_bound = Union[pd.DataFrame, pyspark.sql.DataFrame] if pyspark_exists else pd.DataFrame

DataFrame = TypeVar("DataFrame", bound=df_type_bound)  # type: ignore


class DataFrameDatasetSource(DatasetSource, Generic[DataFrame]):
    def __init__(self, *args: Any, df: DataFrame, dataset_name: str, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.dataset_name = dataset_name
        self._fileobj = self._init_fileobj_from_df(df)
        self.file_size = self._get_size_in_bytes()
        self.file_type = "text/csv"

    @abstractmethod
    def _init_fileobj_from_df(self, df: DataFrame) -> IO[bytes]:
        pass

    def _get_size_in_bytes(self) -> int:
        with self.fileobj() as dataset_file:
            dataset_file.seek(0, os.SEEK_END)
            return dataset_file.tell()

    def get_filename(self) -> str:
        return str(pathlib.Path(self.dataset_name).with_suffix(".csv"))



import io
import os
import time
from typing import Callable, cast, List, Optional, Tuple, Dict, Union, Any
from cleanlab_studio.errors import APIError

import requests
from tqdm import tqdm
import pandas as pd
import numpy as np
import numpy.typing as npt

try:
    import pyspark.sql

    pyspark_exists = True
except ImportError:
    pyspark_exists = False

from cleanlab_studio.internal.types import JSONDict
from cleanlab_studio.version import __version__


base_url = os.environ.get("CLEANLAB_API_BASE_URL", "https://api.cleanlab.ai/api")
cli_base_url = f"{base_url}/cli/v0"
upload_base_url = f"{base_url}/upload/v0"
dataset_base_url = f"{base_url}/datasets"
project_base_url = f"{base_url}/projects"
cleanset_base_url = f"{base_url}/cleansets"
model_base_url = f"{base_url}/v1/deployment"
tlm_base_url = f"{base_url}/v0/trustworthy_llm"


def _construct_headers(
    api_key: Optional[str], content_type: Optional[str] = "application/json"
) -> JSONDict:
    retval = dict()
    if api_key:
        retval["Authorization"] = f"bearer {api_key}"
    if content_type:
        retval["Content-Type"] = content_type
    retval["Client-Type"] = "python-api"
    return retval


def handle_api_error(res: requests.Response) -> None:
    handle_api_error_from_json(res.json())


def handle_api_error_from_json(res_json: JSONDict) -> None:
    if "code" in res_json and "description" in res_json:  # AuthError or UserQuotaError format
        if res_json["code"] == "user_soft_quota_exceeded":
            pass  # soft quota limit is going away soon, so ignore it
        else:
            raise APIError(res_json["description"])
    if res_json.get("error", None) is not None:
        raise APIError(res_json["error"])


def validate_api_key(api_key: str) -> bool:
    res = requests.get(
        cli_base_url + "/validate",
        json=dict(api_key=api_key),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    valid: bool = res.json()["valid"]
    return valid


def is_valid_client_version() -> bool:
    res = requests.post(cli_base_url + "/check_client_version", json=dict(version=__version__))
    handle_api_error(res)
    valid: bool = res.json()["valid"]
    return valid


def initialize_upload(
    api_key: str, filename: str, file_type: str, file_size: int
) -> Tuple[str, List[int], List[str]]:
    res = requests.get(
        f"{upload_base_url}/initialize",
        params=dict(size_in_bytes=str(file_size), filename=filename, file_type=file_type),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    upload_id: str = res.json()["upload_id"]
    part_sizes: List[int] = res.json()["part_sizes"]
    presigned_posts: List[str] = res.json()["presigned_posts"]
    return upload_id, part_sizes, presigned_posts


def complete_file_upload(api_key: str, upload_id: str, upload_parts: List[JSONDict]) -> None:
    request_json = dict(upload_id=upload_id, upload_parts=upload_parts)
    res = requests.post(
        f"{upload_base_url}/complete",
        json=request_json,
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)


def get_proposed_schema(api_key: str, upload_id: str) -> JSONDict:
    res = requests.get(
        f"{upload_base_url}/proposed_schema",
        params=dict(upload_id=upload_id),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    res_json: JSONDict = res.json()
    return res_json


def confirm_schema(
    api_key: str,
    schema: Optional[JSONDict],
    upload_id: str,
) -> None:
    request_json = dict(schema=schema, upload_id=upload_id)
    res = requests.post(
        f"{upload_base_url}/confirm_schema",
        json=request_json,
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)


def get_ingestion_status(api_key: str, upload_id: str) -> JSONDict:
    res = requests.get(
        f"{upload_base_url}/ingestion_status",
        params=dict(upload_id=upload_id),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    res_json: JSONDict = res.json()
    return res_json


def get_dataset_id(api_key: str, upload_id: str) -> JSONDict:
    res = requests.get(
        f"{upload_base_url}/dataset_id",
        params=dict(upload_id=upload_id),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    res_json: JSONDict = res.json()
    return res_json


def get_project_of_cleanset(api_key: str, cleanset_id: str) -> str:
    res = requests.get(
        cli_base_url + f"/cleansets/{cleanset_id}/project",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    project_id: str = res.json()["project_id"]
    return project_id


def get_label_column_of_project(api_key: str, project_id: str) -> str:
    res = requests.get(
        cli_base_url + f"/projects/{project_id}/label_column",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    label_column: str = res.json()["label_column"]
    return label_column


def download_cleanlab_columns(
    api_key: str,
    cleanset_id: str,
    include_cleanlab_columns: bool = True,
    include_project_details: bool = False,
    to_spark: bool = False,
) -> Any:
    """
    Download all rows from specified Cleanlab columns

    :param api_key:
    :param cleanset_id:
    :param include_cleanlab_columns: whether to download all Cleanlab columns or just the clean_label column
    :param include_cleanlab_columns: whether to download columns related to project status such as resolved rows, actions taken, etc.
    :return: return a dataframe, either pandas or spark. Type is Any because don't want to require spark installed
    """
    res = requests.get(
        cli_base_url + f"/cleansets/{cleanset_id}/columns",
        params=dict(
            to_spark=to_spark,
            include_cleanlab_columns=include_cleanlab_columns,
            include_project_details=include_project_details,
        ),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    id_col = get_id_column(api_key, cleanset_id)
    cleanset_json: str = res.json()["cleanset_json"]
    if to_spark:
        if not pyspark_exists:
            raise ImportError(
                "pyspark is not installed. Please install pyspark to download cleanlab columns as a pyspark DataFrame."
            )
        from pyspark.sql import SparkSession

        spark = SparkSession.builder.getOrCreate()
        rdd = spark.sparkContext.parallelize([cleanset_json])
        cleanset_pyspark: pyspark.sql.DataFrame = spark.read.json(rdd)
        cleanset_pyspark = cleanset_pyspark.withColumnRenamed("id", id_col)
        cleanset_pyspark = cleanset_pyspark.sort(id_col)
        return cleanset_pyspark

    cleanset_pd: pd.DataFrame = pd.read_json(cleanset_json, orient="table")
    cleanset_pd.rename(columns={"id": id_col}, inplace=True)
    cleanset_pd.sort_values(by=id_col, inplace=True)
    return cleanset_pd


def download_array(
    api_key: str, cleanset_id: str, name: str
) -> Union[npt.NDArray[np.float_], pd.DataFrame]:
    res = requests.get(
        cli_base_url + f"/cleansets/{cleanset_id}/{name}",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    res_json: JSONDict = res.json()
    if res_json["success"]:
        if res_json["array_type"] == "numpy":
            np_data: npt.NDArray[np.float_] = np.array(res_json[name])
            return np_data
        pd_data: pd.DataFrame = pd.read_json(res_json[name], orient="records")
        return pd_data
    raise APIError(f"{name} for cleanset {cleanset_id} not found")


def get_id_column(api_key: str, cleanset_id: str) -> str:
    res = requests.get(
        cli_base_url + f"/cleansets/{cleanset_id}/id_column",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    id_column: str = res.json()["id_column"]
    return id_column


def get_dataset_of_project(api_key: str, project_id: str) -> str:
    res = requests.get(
        cli_base_url + f"/projects/{project_id}/dataset",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    dataset_id: str = res.json()["dataset_id"]
    return dataset_id


def get_dataset_schema(api_key: str, dataset_id: str) -> JSONDict:
    res = requests.get(
        cli_base_url + f"/datasets/{dataset_id}/schema",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    schema: JSONDict = res.json()["schema"]
    return schema


def get_dataset_details(api_key: str, dataset_id: str) -> JSONDict:
    res = requests.get(
        dataset_base_url + f"/details/{dataset_id}",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    dataset_details: JSONDict = res.json()
    return dataset_details


def clean_dataset(
    api_key: str,
    dataset_id: str,
    project_name: str,
    task_type: str,
    modality: str,
    model_type: str,
    label_column: str,
    feature_columns: List[str],
    text_column: Optional[str],
) -> str:
    request_json = dict(
        name=project_name,
        dataset_id=dataset_id,
        tasktype=task_type,
        modality=modality,
        model_type=model_type,
        label_column=label_column,
        feature_columns=feature_columns,
        text_column=text_column,
    )
    res = requests.post(
        project_base_url + f"/clean",
        headers=_construct_headers(api_key),
        json=request_json,
    )
    handle_api_error(res)
    project_id = res.json()["project_id"]
    return str(project_id)


def get_latest_cleanset_id(api_key: str, project_id: str) -> str:
    res = requests.get(
        cleanset_base_url + f"/project/{project_id}/latest_cleanset_id",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    cleanset_id = res.json()["cleanset_id"]
    return str(cleanset_id)


def get_cleanset_status(api_key: str, cleanset_id: str) -> JSONDict:
    res = requests.get(
        cleanset_base_url + f"/{cleanset_id}/status",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    status: JSONDict = res.json()
    return status


def delete_project(api_key: str, project_id: str) -> None:
    res = requests.delete(project_base_url + f"/{project_id}", headers=_construct_headers(api_key))
    handle_api_error(res)


def poll_progress(
    progress_id: str, request_function: Callable[[str], JSONDict], description: str
) -> JSONDict:
    with tqdm(total=1, desc=description, bar_format="{desc}: {percentage:3.0f}%|{bar}|") as pbar:
        res = request_function(progress_id)
        while res["status"] != "complete":
            if res["status"] == "error":
                raise APIError(res["error_message"])
            pbar.update(float(res["progress"]) - pbar.n)
            time.sleep(0.5)
            res = request_function(progress_id)
        pbar.update(float(1) - pbar.n)
    return res


def upload_predict_batch(api_key: str, model_id: str, batch: io.StringIO) -> str:
    """Uploads prediction batch and returns query ID."""
    url = f"{model_base_url}/{model_id}/upload"
    res = requests.post(
        url,
        headers=_construct_headers(api_key),
    )

    handle_api_error(res)
    presigned_url = res.json()["upload_url"]
    query_id: str = res.json()["query_id"]

    requests.post(presigned_url["url"], data=presigned_url["fields"], files={"file": batch})

    return query_id


def start_prediction(api_key: str, model_id: str, query_id: str) -> None:
    """Starts prediction for query."""
    res = requests.post(
        f"{model_base_url}/{model_id}/predict/{query_id}",
        headers=_construct_headers(api_key),
    )

    handle_api_error(res)


def get_prediction_status(api_key: str, query_id: str) -> Dict[str, str]:
    """Gets status of model prediction query. Returns status, and optionally the result url or error message."""
    res = requests.get(
        f"{model_base_url}/predict/{query_id}",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)

    return cast(Dict[str, str], res.json())


def tlm_prompt(
    api_key: str,
    prompt: str,
    quality_preset: str,
) -> JSONDict:
    """
    Prompt Trustworthy Language Model with a question, and get back its answer along with a confidence score

    Args:
        api_key (str): studio API key for auth
        prompt (str): prompt for TLM to respond to
        quality_preset (str): quality preset to use to generate response

    Returns:
        JSONDict: dictionary with TLM response and confidence score
    """
    res = requests.post(
        f"{tlm_base_url}/prompt",
        json=dict(prompt=prompt, quality=quality_preset),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    return cast(JSONDict, res.json())


def tlm_get_confidence_score(
    api_key: str,
    prompt: str,
    response: str,
    quality_preset: str,
) -> JSONDict:
    """
    Query Trustworthy Language Model for a confidence score for the prompt-response pair.

    Args:
        api_key (str): studio API key for auth
        prompt (str): prompt for TLM to get confidence score for
        response (str): response for TLM to get confidence score for
        quality_preset (str): quality preset to use to generate confidence score

    Returns:
        JSONDict: dictionary with TLM confidence score
    """
    res = requests.post(
        f"{tlm_base_url}/get_confidence_score",
        json=dict(prompt=prompt, response=response, quality=quality_preset),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    return cast(JSONDict, res.json())



"""
Methods for interacting with the command line server API
1:1 mapping with command_line_api.py
"""
import gzip
import json
import os
import pathlib
import asyncio
from typing import List, Any, Optional, Tuple

import aiohttp
import requests
import pandas as pd

from cleanlab_studio.version import __version__
from cleanlab_studio.cli.click_helpers import abort, warn, info
from cleanlab_studio.cli.dataset.image_utils import get_image_filepath
from cleanlab_studio.cli.dataset.schema_types import Schema
from cleanlab_studio.cli.types import JSONDict, IDType, MEDIA_MODALITIES

base_url = os.environ.get("CLEANLAB_API_BASE_URL", "https://api.cleanlab.ai/api")
cli_base_url = f"{base_url}/cli/v0"


MAX_PARALLEL_UPLOADS = 32  # XXX choose this dynamically?
INITIAL_BACKOFF = 0.25  # seconds
MAX_RETRIES = 4


def _construct_headers(
    api_key: Optional[str],
    content_type: Optional[str] = "application/json",
) -> JSONDict:
    retval = dict()
    if api_key:
        retval["Authorization"] = f"bearer {api_key}"
    if content_type:
        retval["Content-Type"] = content_type
    retval["Client-Type"] = "cli"
    retval["User-Agent"] = f"cleanlab-studio/v{__version__}"
    return retval
    return retval


def handle_api_error(res: requests.Response, show_warning: bool = False) -> None:
    handle_api_error_from_json(res.json(), show_warning)


def handle_api_error_from_json(res_json: JSONDict, show_warning: bool = False) -> None:
    if "code" in res_json and "description" in res_json:  # AuthError or UserQuotaError format
        if res_json["code"] == "user_soft_quota_exceeded":
            if show_warning:
                warn(res_json["description"])
        else:
            abort(res_json["description"])
    if res_json.get("error", None) is not None:
        abort(res_json["error"])


def initialize_dataset(api_key: str, schema: Schema) -> str:
    request_json = dict(schema=schema.to_dict())
    res = requests.post(
        cli_base_url + "/datasets", json=request_json, headers=_construct_headers(api_key)
    )
    handle_api_error(res)
    dataset_id: str = res.json()["dataset_id"]
    return dataset_id


def get_existing_ids(api_key: str, dataset_id: str) -> List[IDType]:
    res = requests.get(
        cli_base_url + f"/datasets/{dataset_id}/ids",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    existing_ids: List[IDType] = res.json()["existing_ids"]
    return existing_ids


def get_dataset_schema(api_key: str, dataset_id: str) -> Schema:
    res = requests.get(
        cli_base_url + f"/datasets/{dataset_id}/schema",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    schema = Schema.from_dict(res.json()["schema"])
    return schema


async def upload_rows_async(
    session: aiohttp.ClientSession,
    api_key: str,
    dataset_id: str,
    dataset_filepath: str,
    schema: Schema,
    rows: List[Any],
    filepath_columns: List[str],
    upload_sem: asyncio.Semaphore,
) -> None:
    """
    Upload rows of dataset

    :param session: client session for making http requests
    :param api_key: api key for getting presigned posts
    :param dataset_id: id of dataset to upload files for
    :param dataset_filepath: filepath of dataset to upload files for (can be empty string for simple image upload)
    :param schema: schema of dataset to upload
    :param rows: rows of dataset to upload
    :param filepath_columns: names of any columns containing paths of files that should be uploaded to S3
    :param upload_sem: sempahore to bound number of parallel uploadsb
    """
    modality = schema.metadata.modality
    assert len(filepath_columns) == 0 or modality in MEDIA_MODALITIES
    needs_media_upload = modality in MEDIA_MODALITIES and filepath_columns
    columns = list(schema.fields.keys())

    async with upload_sem:
        if needs_media_upload:
            id_column = schema.metadata.id_column
            assert id_column is not None
            id_column_idx = columns.index(id_column)
            row_ids = [row[id_column_idx] for row in rows]

            upload_tasks = [
                upload_files_for_filepath_column(
                    session=session,
                    api_key=api_key,
                    dataset_id=dataset_id,
                    dataset_filepath=dataset_filepath,
                    rows=rows,
                    row_ids=row_ids,
                    columns=columns,
                    filepath_column=col,
                    media_type=modality.value,
                )
                for col in filepath_columns
            ]
            await asyncio.gather(*upload_tasks)

        url = cli_base_url + f"/datasets/{dataset_id}"
        data = gzip.compress(
            json.dumps(dict(rows=json.dumps(rows), columns=json.dumps(columns))).encode("utf-8")
        )
        headers = _construct_headers(api_key)
        headers["Content-Encoding"] = "gzip"

        async with session.post(url=url, data=data, headers=headers) as res:
            res_text = await res.read()
            handle_api_error_from_json(json.loads(res_text))


async def upload_files_for_filepath_column(
    session: aiohttp.ClientSession,
    api_key: str,
    dataset_id: str,
    dataset_filepath: str,
    rows: List[List[Any]],
    row_ids: List[Any],
    columns: List[str],
    filepath_column: str,
    media_type: str,
) -> None:
    """
    Uploads all files in a filepath column to S3

    :param session: client session for making http requests
    :param api_key: api key for getting presigned posts
    :param dataset_id: id of dataset to upload files for
    :param dataset_filepath: filepath of dataset to upload files for (can be empty string for simple image upload)
    :param rows: dataset rows
    :param row_ids: ids of dataset rows
    :param columns: dataset column names
    :param filepath_column: name of the column to upload files for
    :param media_type: type of media to upload
    """
    sem = asyncio.Semaphore(MAX_PARALLEL_UPLOADS)
    cancelled = asyncio.Event()
    filepath_column_idx = columns.index(filepath_column)
    filepaths: List[str] = [row[filepath_column_idx] for row in rows]
    dataset_dir: pathlib.Path = pathlib.Path(dataset_filepath).parent
    absolute_filepaths = [
        str(get_image_filepath(dataset_dir, row[filepath_column_idx])) for row in rows
    ]
    filepath_to_posts = get_presigned_posts(
        api_key=api_key,
        dataset_id=dataset_id,
        filepaths=filepaths,
        row_ids=row_ids,
        media_type=media_type,
    )
    for coro in asyncio.as_completed(
        [
            post_file(
                session,
                original_filepath,
                absolute_filepath,
                dict(filepath_to_posts[original_filepath]),
                sem,
                cancelled,
            )
            for original_filepath, absolute_filepath in zip(filepaths, absolute_filepaths)
        ]
    ):
        ok, original_filepath = await coro
        if ok:
            info(f"Uploaded {original_filepath}")
        else:
            # cancel tasks we don't need anymore, to give us flexibility to
            # not abort() later but to throw an exception and keep the
            # Python interpreter running without a bunch of leaked
            # coroutines
            cancelled.set()
            abort(f"Failed to upload {original_filepath}")


async def post_file(
    session: aiohttp.ClientSession,
    original_filepath: str,
    absolute_filepath: str,
    post_data: JSONDict,
    sem: asyncio.Semaphore,
    cancelled: asyncio.Event,
) -> Tuple[Optional[bool], str]:
    """
    Upload a single file using a presigned post

    :param session: client session for making http requests
    :original_filepath: the original filepath that appears in the dataset
    :absolute_filepath: the absolute path to the file
    :presigned_post: presigned post to use to upload the file to S3
    :sem: semaphore for parallel uploads
    :cancelled: event that signals if upload should be cancelled
    """
    async with sem:
        presigned_post = post_data["post"]
        # note: we pass in the path and we don't open the file until we
        # get in here, so we don't have tons of concurrently opened
        # files
        if not cancelled.is_set():
            retries = MAX_RETRIES
            wait = INITIAL_BACKOFF
            while retries:
                try:
                    async with session.post(
                        url=presigned_post["url"],
                        data={
                            **presigned_post["fields"],
                            "file": open(absolute_filepath, "rb"),
                        },
                    ) as res:
                        if res.ok:
                            return res.ok, original_filepath
                except Exception:
                    pass  # ignore, will retry
                await asyncio.sleep(wait)
                wait *= 2
                retries -= 1
            return False, original_filepath
    return None, original_filepath


def download_cleanlab_columns(
    api_key: str,
    cleanset_id: str,
    include_cleanlab_columns: bool = False,
    include_project_details: bool = False,
) -> pd.DataFrame:
    """
    Download all rows from specified Cleanlab columns

    :param api_key:
    :param cleanset_id:
    :param all: whether to download all Cleanlab columns or just the clean_label column
    :return: return (rows, id_column)
    """
    res = requests.get(
        cli_base_url
        + f"/cleansets/{cleanset_id}/columns?"
        + f"include_cleanlab_columns={include_cleanlab_columns}"
        + f"&include_project_details={include_project_details}",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    cleanset_json: str = res.json()["cleanset_json"]
    cleanset_df: pd.DataFrame = pd.read_json(cleanset_json, orient="table")
    id_col = get_id_column(api_key, cleanset_id)
    cleanset_df.rename(columns={"id": id_col}, inplace=True)
    return cleanset_df


def get_completion_status(api_key: str, dataset_id: str) -> bool:
    res = requests.get(
        cli_base_url + f"/datasets/{dataset_id}/complete",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    completed: bool = res.json()["complete"]
    return completed


def complete_upload(api_key: str, dataset_id: str) -> None:
    res = requests.patch(
        cli_base_url + f"/datasets/{dataset_id}/complete",
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)


def get_id_column(api_key: str, cleanset_id: str) -> str:
    res = requests.get(
        cli_base_url + f"/cleansets/{cleanset_id}/id_column",
        headers={"Authorization": f"bearer {api_key}"},
    )
    handle_api_error(res)
    id_column: str = res.json()["id_column"]
    return id_column


def validate_api_key(api_key: str) -> bool:
    res = requests.get(
        cli_base_url + "/validate", json=dict(api_key=api_key), headers=_construct_headers(api_key)
    )
    handle_api_error(res)
    valid: bool = res.json()["valid"]
    return valid


def check_client_version() -> bool:
    res = requests.post(cli_base_url + "/check_client_version", json=dict(version=__version__))
    handle_api_error(res)
    valid: bool = res.json()["valid"]
    return valid


def check_dataset_limit(
    api_key: str, file_size: int, modality: str, show_warning: bool = False
) -> JSONDict:
    res = requests.post(
        cli_base_url + "/check_dataset_limit",
        json=dict(file_size=file_size, modality=modality),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res, show_warning=show_warning)
    res_json: JSONDict = res.json()
    return res_json


def get_presigned_posts(
    api_key: str, dataset_id: str, filepaths: List[str], row_ids: List[str], media_type: str
) -> JSONDict:
    res = requests.get(
        cli_base_url + "/media_upload/presigned_posts",
        json=dict(dataset_id=dataset_id, filepaths=filepaths, row_ids=row_ids, type=media_type),
        headers=_construct_headers(api_key),
    )
    handle_api_error(res)
    res_json: JSONDict = res.json()
    return res_json



"""
Helper functions for processing and uploading dataset rows
"""
import asyncio
import decimal
import pathlib
import queue
import re
import threading
from asyncio import Task
from collections import defaultdict
from sys import getsizeof
from typing import (
    Optional,
    Dict,
    IO,
    List,
    Collection,
    Set,
    Any,
    Tuple,
    Coroutine,
    Union,
)

import aiohttp
import click
import pandas as pd
from tqdm import tqdm

from cleanlab_studio.cli import api_service
from cleanlab_studio.cli import click_helpers
from cleanlab_studio.cli.classes.dataset import Dataset
from cleanlab_studio.cli.click_helpers import success, info, progress, abort
from cleanlab_studio.cli.dataset.image_utils import (
    image_file_readable,
    image_file_exists,
    get_image_filepath,
)
from cleanlab_studio.cli.dataset.schema_helpers import get_dataset_filepath_columns
from cleanlab_studio.cli.dataset.schema_types import (
    PYTHON_TYPES_TO_READABLE_STRING,
    DATA_TYPES_TO_PYTHON_TYPES,
    DataType,
    Schema,
    FeatureType,
)
from cleanlab_studio.cli.dataset.upload_types import (
    ValidationWarning,
    WarningLog,
    RowWarningsType,
    warning_to_readable_name,
)
from cleanlab_studio.cli.types import (
    DatasetFileExtension,
    Modality,
    RecordType,
)
from cleanlab_studio.cli.util import (
    is_null_value,
    dump_json,
    init_dataset_from_filepath,
    init_dataset_from_fileobj,
    get_file_size,
)


def get_value_type(val: Any) -> str:
    for python_type, readable_string in PYTHON_TYPES_TO_READABLE_STRING.items():
        if isinstance(val, python_type):
            return readable_string
    return "unrecognized"


def convert_to_python_type(val: Any, data_type: DataType) -> Any:
    if isinstance(val, str):  # int("180.0") gives an error
        if data_type == DataType.integer:
            return int(float(val))
    return DATA_TYPES_TO_PYTHON_TYPES[data_type](val)


def validate_and_process_record(
    record: RecordType,
    schema: Schema,
    seen_ids: Set[str],
    existing_ids: Set[str],
) -> Tuple[Optional[RecordType], Optional[str], Optional[RowWarningsType]]:
    """
    Validate the row against the provided schema; generate warnings where issues are found

    If row ID exists in `existing_ids`, the row has already been uploaded, so we return (None, row ID, None)

    If row ID exists in `seen_ids`, it is a duplicate row, so we return (None, row ID, warnings)

    If row ID is missing, we return (None, None, warnings)

    Otherwise, the only warnings will be for type mismatches and missing values, and we return
    (processed row, row ID, warnings), where warnings is an empty dict if no issues are found.

    :param record: a row in the dataset
    :param schema: dataset schema
    :param seen_ids: the set of row IDs that have been processed so far
    :param existing_ids:
    :return: tuple (processed row: dict[str, any], row ID: optional[str], warnings: dict[warn_type: str, desc: str])
    """
    fields = schema.fields
    id_column = schema.metadata.id_column
    columns = list(fields)
    row_id = record.get(id_column, None)

    if row_id == "" or row_id is None:
        return (
            None,
            None,
            {ValidationWarning.MISSING_ID: [f"Missing ID for record: {dict(record)}."]},
        )

    row_id = str(row_id)
    if row_id in existing_ids:
        return None, row_id, None

    if row_id in seen_ids:
        return (
            None,
            row_id,
            {ValidationWarning.DUPLICATE_ID: [f"Duplicate ID found: {dict(record)}"]},
        )

    warnings: Dict[ValidationWarning, List[str]] = defaultdict(list)

    row = {c: record.get(c, None) for c in columns}

    for column_name, column_value in record.items():
        if column_name not in fields:
            continue
        col_type = fields[column_name].data_type
        col_feature_type = fields[column_name].feature_type

        warning: Optional[Tuple[str, ValidationWarning]] = None
        if is_null_value(column_value):
            row[column_name] = None
            warning = f"{column_name}: value is missing", ValidationWarning.MISSING_VAL
        else:
            if col_feature_type == FeatureType.datetime:
                try:
                    timestamp_value = convert_to_python_type(column_value, col_type)
                    pd.Timestamp(timestamp_value)
                    row[column_name] = timestamp_value
                except (ValueError, TypeError):
                    warning = (
                        f"{column_name}: expected datetime but unable to parse '{column_value}'"
                        f" with {get_value_type(column_value)} type. Datetime strings must be"
                        " parsable by pandas.Timestamp()",
                        ValidationWarning.TYPE_MISMATCH,
                    )
            else:
                if col_type == DataType.string:
                    row[column_name] = str(column_value)  # type coercion
                elif col_type == DataType.integer:
                    if not isinstance(column_value, int):
                        coerced = False
                        if isinstance(column_value, str):
                            if column_value.isdigit():  # e.g. '180'
                                row[column_name] = int(column_value)
                                coerced = True
                            else:
                                try:
                                    temp = float(column_value)
                                    if temp == int(temp):
                                        row[column_name] = int(temp)
                                        coerced = True
                                except ValueError:
                                    pass
                                    # elif
                        elif isinstance(column_value, float) and int(column_value) == column_value:
                            row[column_name] = int(column_value)
                            coerced = True

                        if not coerced:
                            warning = (
                                f"{column_name}: expected 'int' but got '{column_value}' with"
                                f" {get_value_type(column_value)} type",
                                ValidationWarning.TYPE_MISMATCH,
                            )
                elif col_type == DataType.float:
                    if not (isinstance(column_value, int) or isinstance(column_value, float)):
                        coerced = False
                        if isinstance(column_value, str):
                            try:
                                float_value = extract_float_string(column_value)
                                row[column_name] = float(float_value)
                                coerced = True
                            except Exception:
                                pass
                        elif isinstance(column_value, decimal.Decimal):
                            # comes from loading json datasets
                            row[column_name] = float(column_value)
                            coerced = True
                        if not coerced:
                            warning = (
                                f"{column_name}: expected 'float' but got '{column_value}' with"
                                f" {get_value_type(column_value)} type",
                                ValidationWarning.TYPE_MISMATCH,
                            )
                elif col_type == DataType.boolean:
                    if not isinstance(column_value, bool):
                        col_val_lower = str(column_value).lower()
                        if col_val_lower in ["true", "t", "yes", "y", "1", "1.0"]:
                            row[column_name] = True
                        elif col_val_lower in ["false", "f", "no", "n", "0", "0.0"]:
                            row[column_name] = False
                        else:
                            warning = (
                                f"{column_name}: expected 'bool' but got '{column_value}' with"
                                f" {get_value_type(column_value)} type",
                                ValidationWarning.TYPE_MISMATCH,
                            )

        if warning:
            row[column_name] = None  # replace bad value with NULL
            msg, warn_type = warning
            warnings[warn_type].append(msg)
    return row, row_id, warnings


def create_warning_log() -> WarningLog:
    return WarningLog.init()


def update_log_with_warnings(
    log: WarningLog, row_id: Optional[str], warnings: Optional[RowWarningsType]
) -> WarningLog:
    if warnings:
        for warn_type in warnings:
            if warn_type == ValidationWarning.MISSING_ID:
                log_warnings = log.get(warn_type)
                assert isinstance(log_warnings, list)
                log_warnings += warnings[warn_type]
            else:  # ID is present
                assert row_id is not None
                log_warnings = log.get(warn_type)
                assert isinstance(log_warnings, dict)
                log_warnings[row_id] = warnings[warn_type]
    return log


def echo_log_warnings(log: WarningLog) -> None:
    for warning_type in ValidationWarning:
        warning_count = len(log.get(warning_type))
        if warning_count > 0:
            click.echo(f"{warning_to_readable_name(warning_type)}: {warning_count}")


def validate_rows(
    dataset: Union[Dataset[IO[bytes]], Dataset[IO[str]]],
    schema: Schema,
    log: WarningLog,
    upload_queue: "queue.Queue[Optional[List[Any]]]",
    existing_ids: Optional[Collection[str]] = None,
) -> None:
    """Iterates through dataset and validates rows. Places validated rows in upload queue.

    :param dataset_filepath: file path to load dataset from
    :param columns: list of column identifiers for dataset
    :param schema: a validated schema
    :param log: log dict to add warnings to
    :param upload_queue: queue to place validated rows in for upload
    :param existing_ids: set of row IDs that were already uploaded, defaults to None
    """
    existing_ids = set() if existing_ids is None else set([str(x) for x in existing_ids])
    seen_ids: Set[str] = set()

    process_dataset(
        dataset=dataset,
        schema=schema,
        seen_ids=seen_ids,
        existing_ids=existing_ids,
        log=log,
        upload_queue=upload_queue,
    )
    upload_queue.put(None, block=True)


async def upload_rows(
    api_key: str,
    dataset_id: str,
    dataset_filepath: str,
    schema: Schema,
    upload_queue: "queue.Queue[Optional[List[Any]]]",
    rows_per_payload: int,
    filepath_columns: List[str],
) -> None:
    """Gets rows from upload queue and uploads to API.

    :param schema:
    :param dataset_filepath:
    :param api_key: 32-character alphanumeric string
    :param dataset_id: dataset ID
    :param upload_queue: queue to get validated rows from
    :param rows_per_payload: number of rows to upload per payload/chunk
    """
    NUM_PARALLEL_UPLOADS: int = 5
    upload_sem = asyncio.Semaphore(NUM_PARALLEL_UPLOADS)

    async with aiohttp.ClientSession() as session:
        payload = []
        upload_tasks: List[Union[Coroutine[Any, Any, None], Task[None]]] = []
        first_upload = True

        row = upload_queue.get()
        while row is not None:
            payload.append(row)

            if len(payload) >= rows_per_payload:
                upload_tasks.append(
                    asyncio.create_task(
                        api_service.upload_rows_async(
                            session=session,
                            api_key=api_key,
                            dataset_id=dataset_id,
                            dataset_filepath=dataset_filepath,
                            schema=schema,
                            rows=payload,
                            filepath_columns=filepath_columns,
                            upload_sem=upload_sem,
                        )
                    )
                )
                payload = []

                # avoid race condition when creating table
                if first_upload:
                    await upload_tasks[0]
                    upload_tasks = []
                    first_upload = False

            row = upload_queue.get()
            # yield control
            await asyncio.sleep(0)

        # upload remaining rows
        if len(payload) > 0:
            upload_tasks.append(
                api_service.upload_rows_async(
                    session=session,
                    api_key=api_key,
                    dataset_id=dataset_id,
                    dataset_filepath=dataset_filepath,
                    schema=schema,
                    rows=payload,
                    filepath_columns=filepath_columns,
                    upload_sem=upload_sem,
                )
            )

        await asyncio.gather(*upload_tasks)


def check_filepath_column(
    dataset: Union[Dataset[IO[bytes]], Dataset[IO[str]]],
    modality: Modality,
    dataset_filepath: str,
    filepath_column: str,
) -> None:
    """
    Check the filepath column of a dataset to see if any of the filepaths are invalid.
    If >0 filepaths are invalid, print the number of invalid filepaths and prompt user for confirmation about
    outputting filepaths to console.
    """
    dataset_dir = pathlib.Path(dataset_filepath).parent

    if filepath_column not in dataset.get_columns():
        raise ValueError(
            f"No filepath column '{filepath_column}' found in dataset at {dataset_filepath}."
        )
    nonexistent_filepaths = []
    unreadable_filepaths = []
    for record in dataset.read_streaming_records():
        filepath_value = record[filepath_column]
        if not image_file_exists(get_image_filepath(dataset_dir, filepath_value)):
            nonexistent_filepaths.append(filepath_value)
        elif not image_file_readable(get_image_filepath(dataset_dir, filepath_value)):
            unreadable_filepaths.append(filepath_value)

    num_nonexistent_filepaths = len(nonexistent_filepaths)
    num_unreadable_filepaths = len(unreadable_filepaths)
    if num_nonexistent_filepaths + num_unreadable_filepaths > 0:
        click.echo(
            f"Found {num_nonexistent_filepaths} non-existent filepaths and {num_unreadable_filepaths} filepaths that could not be read in specified {modality.value} filepath column: {filepath_column}.\n"
            f"These {num_nonexistent_filepaths + num_unreadable_filepaths} rows with invalid filepaths will be skipped unless the filepaths are corrected. "
            f"Filepaths must be absolute or relative to the directory containing your dataset.\n"
        )
        to_print = click.confirm(
            f"Print the {num_nonexistent_filepaths + num_unreadable_filepaths} filepaths to console?"
        )
        if to_print:
            if num_nonexistent_filepaths > 0:
                click.echo("Non-existent filepaths:\n")
                click.echo(
                    "\n".join(
                        str(get_image_filepath(dataset_dir, pathlib.Path(f)))
                        for f in nonexistent_filepaths
                    )
                )
                click.echo("\n")
            if num_unreadable_filepaths > 0:
                click.echo("Filepaths that could not be read:\n")
                click.echo(
                    "\n".join(
                        str(get_image_filepath(dataset_dir, pathlib.Path(f)))
                        for f in unreadable_filepaths
                    )
                )
                click.echo("\n")

        continue_with_upload = click.confirm(
            "Proceed with dataset upload? (Rows with invalid filepaths will be skipped.)"
        )
        if not continue_with_upload:
            abort("Dataset upload aborted.")


def upload_dataset_from_stream(
    api_key: str,
    dataset_id: str,
    fileobj: Union[IO[str], IO[bytes]],
    schema: Schema,
    dataset_file_extension: DatasetFileExtension,
    existing_ids: Optional[Collection[str]] = None,
    output: Optional[str] = None,
    payload_size: float = 10,
) -> None:
    dataset = init_dataset_from_fileobj(fileobj, dataset_file_extension)
    upload_dataset(
        api_key=api_key,
        dataset_id=dataset_id,
        dataset=dataset,
        schema=schema,
        existing_ids=existing_ids,
        output=output,
        payload_size=payload_size,
    )


def upload_dataset_from_filename(
    api_key: str,
    dataset_id: str,
    filepath: str,
    schema: Schema,
    existing_ids: Optional[Collection[str]] = None,
    output: Optional[str] = None,
    payload_size: float = 10,
) -> None:
    dataset = init_dataset_from_filepath(filepath)
    upload_dataset(
        api_key=api_key,
        dataset_id=dataset_id,
        dataset=dataset,
        schema=schema,
        existing_ids=existing_ids,
        output=output,
        payload_size=payload_size,
        filepath=filepath,
    )


def upload_dataset(
    api_key: str,
    dataset_id: str,
    dataset: Union[Dataset[IO[bytes]], Dataset[IO[str]]],
    schema: Schema,
    existing_ids: Optional[Collection[str]] = None,
    output: Optional[str] = None,
    payload_size: float = 10,
    filepath: str = "",
) -> None:
    """

    :param api_key: 32-character alphanumeric string
    :param dataset_id: dataset ID
    :param filepath: path to dataset file
    :param schema: a validated schema
    :param existing_ids: set of row IDs that were already uploaded
    :param output: filepath to store upload issues in
    :param payload_size: size of each chunk of rows uploaded, in MB
    :return: None
    """
    log = create_warning_log()

    file_size = 0
    if schema.metadata.modality == Modality.image:
        file_size = get_image_dataset_size(dataset=dataset, schema=schema)
        if filepath:
            get_file_size(filepath)
    else:
        file_size = get_file_size(filepath)

    api_service.check_dataset_limit(
        api_key=api_key,
        file_size=file_size,
        modality=schema.metadata.modality.value,
        show_warning=False,
    )

    # NOTE: makes simplifying assumption that first row size is representative of all row sizes
    row_size = getsizeof(next(dataset.read_streaming_records()))
    rows_per_payload = int(payload_size * 1e6 / row_size)
    upload_queue: queue.Queue[Optional[List[Any]]] = queue.Queue(maxsize=2 * rows_per_payload)

    # create validation process
    validation_thread = threading.Thread(
        target=validate_rows,
        kwargs={
            "dataset": dataset,
            "schema": schema,
            "log": log,
            "upload_queue": upload_queue,
            "existing_ids": existing_ids,
        },
    )

    filepath_columns = get_dataset_filepath_columns(dataset, schema)

    # start and join processes
    validation_thread.start()
    asyncio.run(
        upload_rows(
            api_key=api_key,
            dataset_id=dataset_id,
            dataset_filepath=filepath,
            schema=schema,
            upload_queue=upload_queue,
            rows_per_payload=rows_per_payload,
            filepath_columns=filepath_columns,
        )
    )
    validation_thread.join()

    # Check against soft quota, warn if applicable
    api_service.check_dataset_limit(
        api_key=api_key, file_size=0, modality=schema.metadata.modality.value, show_warning=True
    )

    total_warnings: int = sum([len(log.get(w)) for w in ValidationWarning])
    issues_found = total_warnings > 0

    conclude_upload = True
    if not issues_found:
        success("\nNo issues were encountered when uploading your dataset. Nice!")
        api_service.complete_upload(api_key=api_key, dataset_id=dataset_id)
    else:
        info(f"\n{total_warnings} issues were encountered when uploading your dataset.")
        echo_log_warnings(log)

        if schema.metadata.modality == Modality.image:
            num_invalid_filepaths = len(log.get(ValidationWarning.MISSING_FILE)) + len(
                log.get(ValidationWarning.UNREADABLE_FILE)
            )

            if num_invalid_filepaths > 0:
                click_helpers.warn(
                    f"{num_invalid_filepaths} rows failed to upload due to invalid filepaths."
                )
                conclude_upload = click.confirm(
                    f"Conclude dataset upload? Concluding means that no more rows can be uploaded. "
                    f"Otherwise, submit 'n' and resume uploading later with the dataset ID ({dataset_id}) after fixing rows with invalid filepaths."
                )
                if not conclude_upload:
                    click_helpers.warn("Dataset upload is incomplete.")

        if not output:
            output = click_helpers.confirm_save_prompt_filepath(
                save_message="Save the issues for viewing?",
                save_default=None,
                prompt_message=(
                    "Specify a filename for the dataset issues. Leave this blank to use default"
                ),
                prompt_default="issues.json",
                no_save_message="Dataset type issues were not saved.",
            )
        # if we have an output after the above prompt (or originally provided)
        if output:
            save_warning_log(log, output)
            click_helpers.confirm_open_file("Open your issues file for viewing?", filepath=output)

    if conclude_upload:
        api_service.complete_upload(api_key=api_key, dataset_id=dataset_id)
        click_helpers.success(
            "Upload completed. View your uploaded dataset at https://app.cleanlab.ai"
        )


def group_feature_types(schema: Schema) -> Dict[FeatureType, List[str]]:
    """
    Given a schema, return a dict mapping each feature type to the list of columns with said feature type
    """
    feature_types_to_columns = defaultdict(list)
    for field_name, spec in schema.fields.items():
        feature_types_to_columns[spec.feature_type].append(field_name)
    return feature_types_to_columns


def save_warning_log(warning_log: WarningLog, filename: str) -> None:
    if not filename:
        raise ValueError("No filepath provided for saving warning_log")
    progress(f"Writing issues to {filename}...")
    dump_json(filename, warning_log.to_dict(readable=True))
    success("Saved.\n")


def extract_float_string(column_value: str) -> str:
    """
    Floating point: Decimal number containing a decimal point, optionally preceded by a + or - sign,
    and then optionally preceded by a currency symbol.
    Optionally followed by the e or E character and a decimal number, and optionally ended with a % symbol.

    The extracted string should include only the value (i.e. includes decimals and E notation), and no other symbols.
    If a - sign precedes the value string, it is preserved in the return value.

    Reference: https://docs.python.org/3/library/re.html#simulating-scanf
    """
    float_regex_pattern = r"^([+-]?)[$€£]?((\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?)[%]?$"
    match = re.search(float_regex_pattern, column_value)
    if match:
        group_1 = match.group(1)
        if group_1 in ["+", "-"]:
            if group_1 == "-":
                return f"-{match.group(2)}"
            else:
                return match.group(2)
        else:
            return match.group(2)
    else:
        return ""


def process_dataset(
    dataset: Union[Dataset[IO[str]], Dataset[IO[bytes]]],
    schema: Schema,
    seen_ids: Set[str],
    existing_ids: Set[str],
    log: WarningLog,
    upload_queue: Optional["queue.Queue[Optional[List[Any]]]"] = None,
) -> None:
    """
    Validate and processes records, while updating the warning log
    If an upload queue is provided, valid rows are put on it
    """
    for record in tqdm(
        dataset.read_streaming_records(), total=len(dataset), initial=0, leave=True, unit=" rows"
    ):
        row, row_id, warnings = validate_and_process_record(record, schema, seen_ids, existing_ids)
        update_log_with_warnings(log, row_id, warnings)
        # row and row ID both present, i.e. row will be uploaded
        if row_id:
            seen_ids.add(row_id)

        if upload_queue and row is not None:
            upload_queue.put(list(row.values()), block=True)


def get_image_dataset_size(
    dataset: Union[Dataset[IO[bytes]], Dataset[IO[str]]], schema: Schema
) -> int:
    """Returns total image dataset size by summing file sizes of each image"""
    filepath_columns = get_dataset_filepath_columns(dataset, schema)
    return sum(
        sum(
            [
                get_file_size(filepath=record[filepath_column], ignore_missing_files=True)
                for record in dataset.read_streaming_records()
            ]
        )
        for filepath_column in filepath_columns
    )



import json
import os
from typing import Optional
import semver

from cleanlab_studio.version import SETTINGS_VERSION, MIN_SETTINGS_VERSION, MAX_SETTINGS_VERSION
from cleanlab_studio.internal.types import CleanlabSettingsDict


class CleanlabSettings:
    def __init__(self, version: Optional[str], api_key: Optional[str]):
        self.version = version
        self.api_key = api_key

    @staticmethod
    def init_base() -> "CleanlabSettings":
        return CleanlabSettings(version=SETTINGS_VERSION, api_key=None)

    @staticmethod
    def from_dict(d: CleanlabSettingsDict) -> "CleanlabSettings":
        return CleanlabSettings(
            version=d.get("version", None),
            api_key=d.get("api_key", None),
        )

    def to_dict(self) -> CleanlabSettingsDict:
        return dict(version=self.version, api_key=self.api_key)

    @staticmethod
    def get_cleanlab_dir() -> str:
        XDG_CONFIG_HOME = os.environ.get("XDG_CONFIG_HOME")
        CONFIG_HOME = XDG_CONFIG_HOME if XDG_CONFIG_HOME else "~/.config"
        cleanlab_dir = os.path.expanduser(os.path.join(CONFIG_HOME, "cleanlab"))
        return cleanlab_dir

    @staticmethod
    def get_settings_filepath() -> str:
        return os.path.join(CleanlabSettings.get_cleanlab_dir(), "settings.json")

    @staticmethod
    def init_cleanlab_dir() -> None:
        """
        Initializes a cleanlab config directory if one does not currently exist. No-op if one already exists.
        :return:
        """
        cleanlab_dir = CleanlabSettings.get_cleanlab_dir()
        os.makedirs(cleanlab_dir, exist_ok=True)

    @staticmethod
    def init_cleanlab_settings() -> None:
        """
        Initializes and saves a base Cleanlab settings file. No-op if one already exists.
        :return:
        """
        CleanlabSettings.init_cleanlab_dir()
        settings_filepath = CleanlabSettings.get_settings_filepath()
        if not os.path.exists(settings_filepath):
            base = CleanlabSettings.init_base()
            base.save()

    @staticmethod
    def load() -> "CleanlabSettings":
        filepath = CleanlabSettings.get_settings_filepath()
        with open(filepath, "r") as f:
            settings_dict = json.load(f)
        settings = CleanlabSettings.from_dict(settings_dict)
        settings.validate_version()
        return settings

    def update_version(self) -> None:
        self.version = SETTINGS_VERSION
        self.save()

    def validate_version(self) -> None:
        if semver.compare(MIN_SETTINGS_VERSION, self.version) == 1:
            # TODO add proper settings migrations
            raise ValueError("Settings file must be migrated or re-generated.")
        elif semver.compare(MAX_SETTINGS_VERSION, self.version) == -1:
            raise ValueError(
                "CLI is not up to date with your settings version. Run 'pip install --upgrade cleanlab-studio'."
            )

    def save(self) -> None:
        with open(CleanlabSettings.get_settings_filepath(), "w") as f:
            json.dump(self.to_dict(), f)


def get_cleanlab_settings() -> "CleanlabSettings":
    settings_filepath = CleanlabSettings.get_settings_filepath()
    if not os.path.exists(settings_filepath):
        CleanlabSettings.init_cleanlab_settings()

    return CleanlabSettings.load()



import io
from typing import IO
import pandas as pd

from .dataframe_dataset_source import DataFrameDatasetSource


class PandasDatasetSource(DataFrameDatasetSource[pd.DataFrame]):
    def _init_fileobj_from_df(self, df: pd.DataFrame) -> IO[bytes]:
        fileobj = io.BytesIO()
        df.to_csv(fileobj, index=False)
        fileobj.seek(0)
        return fileobj



import click
from cleanlab_studio.cli.cleanset.download import download


@click.group(help="download Cleanlab columns from cleansets")
def cleanset() -> None:
    pass


cleanset.add_command(download)



import json
from cleanlab_studio.cli.api_service import validate_api_key
from cleanlab_studio.cli.click_helpers import *
from cleanlab_studio.cli.decorators.previous_state import PreviousState
from cleanlab_studio.cli.decorators import previous_state
from cleanlab_studio.internal.settings import CleanlabSettings
import click


@click.command(help="authentication for Cleanlab Studio")
@click.option(
    "--key",
    "--k",
    type=str,
    prompt=True,
    help="API key for CLI uploads. You can get this from https://app.cleanlab.ai/upload.",
)
@previous_state
def login(prev_state: PreviousState, key: str) -> None:
    prev_state.init_state(dict(command="login", args=dict(key=key)))
    CleanlabSettings.init_cleanlab_dir()

    # validate API key
    valid_key = validate_api_key(key)
    if not valid_key:
        abort("API key is invalid. Check https://app.cleanlab.ai/upload for your current API key.")

    # save API key
    try:
        settings = CleanlabSettings.load()
        settings.api_key = key
        settings.save()

    except json.decoder.JSONDecodeError:
        error("CLI settings are corrupted and could not be read.")
        overwrite = click.confirm(
            "Create a new settings file with the provided API key?",
            default=None,
        )
        if overwrite:
            base = CleanlabSettings.init_base()
            base.api_key = key
            base.save()
        else:
            abort(
                "Settings file is corrupted, unable to login. "
                "Either re-run this command and allow a new settings file to be created, "
                f"or manually fix the settings file at {CleanlabSettings.get_settings_filepath()}"
            )

    success(f"API key is valid. API key stored in {CleanlabSettings.get_cleanlab_dir()}")



import pathlib
from typing import List

import pandas as pd
import pytest

from cleanlab_studio.cli.classes.excel_dataset import ExcelDataset
from .constants import DATASETS_DIR
from .utils import assert_dataset_matches_dataframe, excel_filepath_to_dataframe


EXCEL_DATASETS: List[pathlib.Path] = [*DATASETS_DIR.glob(r"*.xlsx")]


@pytest.mark.parametrize(
    ("excel_filepath", "excel_df"),
    (
        (excel_filepath, excel_filepath_to_dataframe(excel_filepath))
        for excel_filepath in EXCEL_DATASETS
    ),
)
def test_excel_dataset_from_filepath_matches_expected(
    excel_filepath: pathlib.Path, excel_df: pd.DataFrame
):
    """Tests that ExcelDataset, loaded from filepath, matches excel loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load ExcelDataset from filepath
    excel_dataset = ExcelDataset(filepath=str(excel_filepath), file_type="xlsx")

    # check that methods match values read by pandas
    assert_dataset_matches_dataframe(excel_dataset, excel_df)


@pytest.mark.parametrize(
    ("excel_filepath", "excel_df"),
    (
        (excel_filepath, excel_filepath_to_dataframe(excel_filepath))
        for excel_filepath in EXCEL_DATASETS
    ),
)
def test_excel_dataset_from_fileobj_matches_expected(
    excel_filepath: pathlib.Path, excel_df: pd.DataFrame
):
    """Tests that excelDataset, loaded from file object, matches excel loaded by Pandas.

    Checks:
    - __len__
    - get_columns
    - read_streaming_records
    - read_streaming_values
    - read_file_as_dataframe
    """
    # load excelDataset from object
    with open(excel_filepath, **ExcelDataset.READ_ARGS) as excel_file:
        excel_dataset = ExcelDataset(fileobj=excel_file, file_type="xlsx")

        # check that methods match values read by pandas
        assert_dataset_matches_dataframe(excel_dataset, excel_df)



from os import path

from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Get version number and store it in __version__
exec(open("cleanlab_studio/version.py").read())

setup(
    name="cleanlab-studio",
    version=__version__,
    license="MIT",
    author="Cleanlab Inc",
    author_email="team@cleanlab.ai",
    description="Client interface for all things Cleanlab Studio",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cleanlab/cleanlab-studio",
    project_urls={
        "Bug Tracker": "https://github.com/cleanlab/cleanlab-studio/issues",
        "Documentation": "https://cleanlab-studio.readthedocs.io/en/latest/index.html",
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: OS Independent",
    ],
    keywords="cleanlab",
    packages=find_packages(exclude=[]),
    py_modules=["main"],
    python_requires=">=3.8",
    install_requires=[
        "aiohttp>=3.8.1",
        "Click>=8.1.0,<=8.1.3",
        "colorama>=0.4.4",
        "pandas==2.*",
        "pyexcel>=0.7.0",
        "pyexcel-xls>=0.7.0",
        "pyexcel-xlsx>=0.6.0",
        "requests>=2.27.1",
        "tqdm>=4.64.0",
        "ijson>=3.1.4",
        "jsonstreams>=0.6.0",
        "semver>=2.13.0,<3.0.0",
        "Pillow>=9.2.0",
        "typing_extensions==4.2.0",
        "openpyxl==3.0.10",
        "validators>=0.20.0",
    ],
    extras_require={
        "docs": [
            "sphinx==6.2.1",
            "sphinx_toolbox==3.4.0",
            "furo==2022.12.07",
            "nbsphinx==0.9.1",
        ],
    },
    entry_points="""
        [console_scripts]
        cleanlab=cleanlab_studio.cli:main
    """,
)



import mimetypes
import pathlib
from typing import Any, Optional

from .dataset_source import DatasetSource


class FilepathDatasetSource(DatasetSource):
    def __init__(
        self,
        *args: Any,
        filepath: pathlib.Path,
        dataset_name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.dataset_name = dataset_name if dataset_name is not None else filepath.name
        self.file_size = filepath.stat().st_size
        maybe_file_type = mimetypes.guess_type(filepath)[0]
        if maybe_file_type is None:
            raise ValueError(
                f"Could not identify type of file at {filepath}. Make sure file name has valid extension"
            )
        self.file_type = maybe_file_type
        self._filepath = filepath

    def get_filename(self) -> str:
        assert self._filepath is not None
        return self._filepath.name



import click
from cleanlab_studio.cli.click_helpers import abort
from cleanlab_studio.internal.api import validate_api_key
from cleanlab_studio.internal.settings import CleanlabSettings
from typing import Optional


class AuthConfig:
    def __init__(self) -> None:
        self.api_key: Optional[str] = None

    def get_api_key(self) -> str:
        if self.api_key is None:
            try:
                api_key = CleanlabSettings.load().api_key
                if api_key is None or not validate_api_key(api_key):
                    raise ValueError("Invalid API key.")
                else:
                    self.api_key = api_key
                    return api_key
            except (FileNotFoundError, KeyError, ValueError):
                abort("No valid API key found. Run 'cleanlab login' before running this command.")
        assert self.api_key is not None
        return self.api_key


auth_config = click.make_pass_decorator(AuthConfig, ensure=True)



import click
import os
import subprocess

from cleanlab_studio.internal.api import api
from cleanlab_studio.cli.click_helpers import abort
from cleanlab_studio.cli.dataset.commands import dataset
from cleanlab_studio.cli.cleanset.commands import cleanset
from cleanlab_studio.cli.login.login import login
from cleanlab_studio.internal.settings import CleanlabSettings
from cleanlab_studio.version import __version__


@click.group()
@click.pass_context
def cli(ctx: click.Context) -> None:
    if ctx.invoked_subcommand == "version":
        return  # avoid RTT / dependence on API to get client version
    CleanlabSettings.init_cleanlab_settings()
    valid_version = api.is_valid_client_version()
    if not valid_version:
        abort(
            "CLI is out of date and must be updated. Run 'pip install --upgrade cleanlab-studio'."
        )
    pass


@click.command(name="version", help="get version of cleanlab-studio")
def version() -> None:
    # detect if user has installed the package from source in editable mode
    try:
        dir = os.path.dirname(__file__)
        sha = subprocess.check_output(
            ["git", "rev-parse", "HEAD"], cwd=dir, stderr=subprocess.DEVNULL
        )[:10].decode("utf8")
        dirty = subprocess.check_output(["git", "diff", "--stat"], stderr=subprocess.DEVNULL) != b""
        print(f"cleanlab-studio {__version__} (git sha1 {sha}{'-dirty' if dirty else ''})")
    except:
        print(f"cleanlab-studio {__version__}")


cli.add_command(login)
cli.add_command(dataset)
cli.add_command(cleanset)
cli.add_command(version)



from typing import Any, Dict, Optional, TypedDict


JSONDict = Dict[str, Any]
FieldSchemaDict = Dict[str, Dict[str, Any]]


class CleanlabSettingsDict(TypedDict):
    version: Optional[str]
    api_key: Optional[str]



"""
Helper functions for working with schemas
"""
import decimal
import json
import pathlib
import random
import re
from typing import Any, Collection, Dict, List, Optional, Sized, Tuple, IO, Union

import pandas as pd
import requests
import semver
import validators
from pandas import NaT

from cleanlab_studio.cli.classes.dataset import Dataset
from cleanlab_studio.cli.click_helpers import abort, info, progress, success
from cleanlab_studio.cli.dataset.schema_types import (
    DataType,
    FeatureType,
    MEDIA_FEATURE_TYPES,
    Schema,
)
from cleanlab_studio.cli.types import Modality
from cleanlab_studio.cli.util import dump_json
from cleanlab_studio.errors import ColumnMismatchError, EmptyDatasetError
from cleanlab_studio.internal.types import JSONDict
from cleanlab_studio.version import MAX_SCHEMA_VERSION, MIN_SCHEMA_VERSION, SCHEMA_VERSION


def _find_best_matching_column(target_column: str, columns: List[str]) -> Optional[str]:
    """
    Find the column from `columns` that is the closest match to the `target_col`.
    If no columns are likely, pick the first column of `columns`
    If no columns are provided, return None

    :param target_column: some reserved column name, typically: 'id', 'label', or 'text'
    :param columns: non-empty list of column names
    :return:
    """
    if len(columns) == 0:
        return None

    if target_column == "id":
        regex = r"id"
    elif target_column == "filepath":
        regex = r"file|path|dir"
    else:
        regex = r""

    poss = []
    for c in columns:
        if c.lower() == target_column:
            return c
        elif any(re.findall(regex, c, re.IGNORECASE)):
            poss.append(c)

    if len(poss) > 0:  # pick first possibility
        return poss[0]
    else:
        return columns[0]


def load_schema(filepath: str) -> JSONDict:
    with open(filepath, "r") as f:
        schema_dict: JSONDict = json.load(f)
        return schema_dict


def validate_schema(schema: Schema) -> None:
    """
    Checks that:
    (1) all schema column names are strings
    (2) all schema column types are recognized
    Note that schema initialization already checks that all keys are present and that fields are valid.
    :param schema:
    :return: raises a ValueError if any checks fail
    """

    # check schema version validity
    schema_version = schema.version
    if (
        semver.VersionInfo.parse(MIN_SCHEMA_VERSION).compare(schema_version) == 1
    ):  # min schema > schema_version
        raise ValueError(
            "This schema version is incompatible with this version of the CLI. "
            "A new schema should be generated using 'cleanlab dataset schema generate'"
        )
    elif semver.VersionInfo.parse(MAX_SCHEMA_VERSION).compare(schema_version) == -1:
        raise ValueError(
            "CLI is not up to date with your schema version. Run 'pip install --upgrade cleanlab-studio'."
        )

    metadata = schema.metadata

    # Advanced validation checks: this should be aligned with ConfirmSchema's validate() function
    ## Check that specified ID column has the feature_type 'identifier'
    id_column_name = metadata.id_column
    id_column_spec_feature_type = schema.fields[id_column_name].feature_type
    if id_column_spec_feature_type != FeatureType.identifier:
        raise ValueError(
            f"ID column field {id_column_name} must have feature type: 'identifier', but has"
            f" feature type: '{id_column_spec_feature_type}'"
        )

    ## Check that there exists at least one categorical column (to be used as label)
    has_categorical = any(
        spec.feature_type == FeatureType.categorical for spec in schema.fields.values()
    )
    if not has_categorical:
        raise ValueError(
            "Dataset does not seem to contain a label column. (None of the fields is categorical.)"
        )

    ## If tabular modality, check that there are at least two variable (i.e. categorical, numeric, datetime) columns
    modality = metadata.modality
    variable_fields = {FeatureType.categorical, FeatureType.numeric, FeatureType.datetime}
    if modality == Modality.tabular:
        num_variable_columns = sum(
            int(spec.feature_type in variable_fields) for spec in schema.fields.values()
        )
        if num_variable_columns < 2:
            raise ValueError(
                "Dataset modality is tabular; there must be at least one categorical field and one"
                " other variable field (i.e. categorical, numeric, or datetime)."
            )

    ## If text modality, check that at least one column has feature type 'text'
    elif modality == Modality.text:
        has_text = any(spec.feature_type == FeatureType.text for spec in schema.fields.values())
        if not has_text:
            raise ValueError("Dataset modality is text, but none of the fields is a text column.")

    elif modality == Modality.image:
        image_columns = [
            col for col, spec in schema.fields.items() if spec.feature_type == FeatureType.image
        ]
        if not image_columns:
            raise ValueError(
                "Dataset modality is image, but none of the fields is an image column."
            )
        if len(image_columns) > 1:
            raise ValueError("More than one image column in a dataset is not currently supported.")


def multiple_separate_words_detected(values: Collection[Any]) -> bool:
    avg_num_words = sum([len(str(v).split()) for v in values]) / len(values)
    return avg_num_words >= 3


def is_filepath(string: str, check_existing: bool = False) -> bool:
    if pathlib.Path(string).suffix == "" or " " in string:
        return False
    if check_existing:
        return pathlib.Path(string).exists()
    return True


def is_url(string: str, check_existing: bool = False) -> bool:
    if not validators.url(string):
        return False

    try:
        if check_existing:
            requests.head(string)
            return True
    except requests.RequestException:
        return False

    return True


def get_validation_sample_size(values: Sized) -> int:
    return min(20, len(values))


def string_values_are_datetime(values: Collection[Any]) -> bool:
    try:
        # check for datetime first
        val_sample = random.sample(list(values), get_validation_sample_size(values))
        for s in val_sample:
            res = pd.to_datetime(s)
            if res is NaT:
                raise ValueError
    except Exception:
        return False
    return True


def string_values_are_integers(values: Collection[Any]) -> bool:
    try:
        val_sample = random.sample(list(values), get_validation_sample_size(values))
        for s in val_sample:
            if str(int(s)) != s:
                return False
    except Exception:
        return False
    return True


def string_values_are_floats(values: Collection[Any]) -> bool:
    try:
        val_sample = random.sample(list(values), get_validation_sample_size(values))
        for s in val_sample:
            float(s)
    except Exception:
        return False
    return True


def values_are_filepaths(values: Collection[Any]) -> bool:
    val_sample = random.sample(list(values), get_validation_sample_size(values))
    for s in val_sample:
        if not is_filepath(s):
            return False
    return True


def values_are_urls(values: Collection[Any]) -> bool:
    val_sample = random.sample(list(values), get_validation_sample_size(values))
    return all(is_url(s) for s in val_sample)


def infer_types(values: Collection[Any]) -> Tuple[DataType, FeatureType]:
    """
    Infer the data type and feature type of a collection of a values using simple heuristics.

    :param values: a Collection of data values (that are not null and not empty string)
    """
    counts = {DataType.string: 0, DataType.integer: 0, DataType.float: 0, DataType.boolean: 0}
    ID_RATIO_THRESHOLD = 0.97  # lowerbound
    CATEGORICAL_RATIO_THRESHOLD = 0.20  # upperbound

    ratio_unique = len(set(values)) / len(values)
    for v in values:
        if v == "":
            continue
        if isinstance(v, str):
            counts[DataType.string] += 1
        elif isinstance(v, float):
            counts[DataType.float] += 1
        elif isinstance(v, bool):  # must come before int: isinstance(True, int) evaluates to True
            counts[DataType.boolean] += 1
        elif isinstance(v, int):
            counts[DataType.integer] += 1
        elif isinstance(v, decimal.Decimal):  # loading from JSONs can produce Decimal values
            counts[DataType.float] += 1
        else:
            raise ValueError(f"Value {v} has an unrecognized type: {type(v)}")

    ratios: Dict[DataType, float] = {k: v / len(values) for k, v in counts.items()}
    max_count_type = max(ratios, key=lambda k: ratios[k])

    # preliminary check: ints/floats may be loaded as strings
    if max_count_type == DataType.string:
        if string_values_are_integers(values):
            max_count_type = DataType.integer
        elif string_values_are_floats(values):
            max_count_type = DataType.float

    if max_count_type == DataType.string:
        if string_values_are_datetime(values):
            return DataType.string, FeatureType.datetime
        # is string type
        if ratio_unique >= ID_RATIO_THRESHOLD:
            # almost all unique values, i.e. either ID, text
            if multiple_separate_words_detected(values):
                return DataType.string, FeatureType.text
            else:
                if values_are_urls(values) or values_are_filepaths(values):
                    return DataType.string, FeatureType.image
                return DataType.string, FeatureType.identifier
        elif ratio_unique <= CATEGORICAL_RATIO_THRESHOLD:
            return DataType.string, FeatureType.categorical
        else:
            return DataType.string, FeatureType.text

    elif max_count_type == DataType.integer:
        if ratio_unique >= ID_RATIO_THRESHOLD:
            return DataType.integer, FeatureType.identifier
        elif ratio_unique <= CATEGORICAL_RATIO_THRESHOLD:
            return DataType.integer, FeatureType.categorical
        else:
            return DataType.integer, FeatureType.numeric
    elif max_count_type == DataType.float:
        return DataType.float, FeatureType.numeric
    elif max_count_type == DataType.boolean:
        return DataType.boolean, FeatureType.boolean
    else:
        return DataType.string, FeatureType.text


def propose_schema(
    dataset: Union[Dataset[IO[str]], Dataset[IO[bytes]]],
    name: str,
    columns: Optional[Collection[str]] = None,
    id_column: Optional[str] = None,
    modality: Optional[str] = None,
    sample_size: int = 10000,
    max_rows_checked: int = 200000,
) -> Schema:
    """
    Generates a schema for a dataset based on a sample of the dataset's rows.

    Dataset columns with no name will not be included in the schema.

    :param dataset:
    :param name: name of dataset
    :param columns: columns to generate a schema for
    :param id_column: ID column name
    :param filepath_column: filepath column name, i.e. name of column holding media filepaths (needed if modality is image)
    :param modality: data modality
    :param sample_size: default of 1000
    :param max_rows_checked: max rows to sample from
    :return:

    """
    # The arguments are intended to be required for the command-line interface, but are optional for Cleanlab Studio.
    if len(dataset) < 1:
        raise EmptyDatasetError("Cannot propose schema for empty dataset.")

    # fill optional arguments if necessary
    if columns is None:
        columns = dataset.get_columns()

    if modality is None:
        # suggested modality can be set to a media modality on line 382
        if len(columns) > 5:
            modality = Modality.tabular.value
        else:
            modality = Modality.text.value

    # dataset = []
    rows = []
    for idx, row in enumerate(dataset.read_streaming_values()):
        if idx >= max_rows_checked:
            break
        if idx < sample_size:
            rows.append(row)
        else:
            random_idx = random.randint(0, idx)
            if random_idx < sample_size:
                rows[random_idx] = row

    try:
        df = pd.DataFrame(data=rows, columns=list(columns))
    except ValueError as e:
        raise ColumnMismatchError(str(e))

    schema_dict = dict()
    fields_dict = dict()
    for column_name in columns:
        if column_name == "":
            continue
        column_values = list(df[column_name][~df[column_name].isna()])
        column_values = [v for v in column_values if v != ""]

        if len(column_values) == 0:  # all values in column are empty, give default string[text]
            fields_dict[column_name] = dict(
                data_type=DataType.string.value, feature_type=FeatureType.text.value
            )
            continue

        col_data_type, col_feature_type = infer_types(column_values)
        fields_dict[column_name] = dict(
            data_type=col_data_type.value, feature_type=col_feature_type.value
        )
        if col_feature_type in MEDIA_FEATURE_TYPES:
            modality = col_feature_type.value

    schema_dict["fields"] = fields_dict

    if id_column is None:
        id_columns = [
            k
            for k, spec in schema_dict["fields"].items()
            if spec["feature_type"] == FeatureType.identifier.value
        ]
        if len(id_columns) == 0:
            id_columns = list(columns)
        id_column = _find_best_matching_column("id", id_columns)
    else:
        if id_column not in columns:
            abort(f"ID column '{id_column}' does not exist in the dataset.")

    assert id_column is not None

    metadata: Dict[str, Optional[str]] = dict(id_column=id_column, modality=modality, name=name)
    return Schema.create(metadata=metadata, fields=fields_dict, version=SCHEMA_VERSION)


def save_schema(schema: JSONDict, filename: Optional[str]) -> None:
    """

    :param schema:
    :param filename: filename to save schema with
    :return:
    """
    if filename == "":
        filename = "schema.json"
    if filename:
        progress(f"Writing schema to {filename}...")
        dump_json(filename, schema)
        success("Saved.")
    else:
        info("Schema was not saved.")


def get_dataset_filepath_columns(
    dataset: Union[Dataset[IO[bytes]], Dataset[IO[str]]], schema: Schema
) -> List[str]:
    media_columns = [
        field_name
        for field_name, field_spec in schema.fields.items()
        if field_spec.feature_type in MEDIA_FEATURE_TYPES
    ]
    df = dataset.read_file_as_dataframe()
    return [
        col
        for col in media_columns
        if not values_are_urls(df[col]) and values_are_filepaths(df[col])
    ]



from typing import Generator, List, Any, Dict, IO

import pandas as pd
import ijson

from .dataset import Dataset
from cleanlab_studio.cli.types import RecordType


class JsonDataset(Dataset[IO[bytes]]):
    READ_ARGS: Dict[str, str] = {"mode": "rb"}

    def read_streaming_records(self) -> Generator[RecordType, None, None]:
        with self.fileobj() as f:
            for r in ijson.items(f, "item"):
                yield r

    def read_streaming_values(self) -> Generator[List[Any], None, None]:
        with self.fileobj() as f:
            for r in ijson.items(f, "item"):
                yield r.values()

    def read_file_as_dataframe(self) -> pd.DataFrame:
        with self.fileobj() as f:
            df = pd.read_json(f, orient="records", convert_axes=False, convert_dates=False)

        df.index = df.index.astype("str")
        df["id"] = df.index
        return df



from abc import abstractmethod
import contextlib
from typing import Optional, List, IO, Generator, Any, Iterator, Dict, Generic, TypeVar, Union

import pandas as pd

from cleanlab_studio.cli.types import RecordType


FileObj = TypeVar("FileObj", bound=Union[IO[str], IO[bytes]])


class Dataset(Generic[FileObj]):
    READ_ARGS: Dict[str, str] = {}

    def __init__(self, filepath: Optional[str] = None, fileobj: Optional[FileObj] = None):
        if filepath is None and fileobj is None:
            raise ValueError(
                "One of `filepath` or `fileobj` must be provided to initialize `Dataset`."
            )

        self._filepath: Optional[str] = filepath
        self._fileobj: Optional[FileObj] = fileobj

        self._num_rows: Optional[int] = None

    @contextlib.contextmanager
    def fileobj(self) -> Iterator[FileObj]:
        """Yields open IO object to dataset file."""
        if self._filepath is not None:
            with open(self._filepath, **self.READ_ARGS) as dataset_file:  # type: ignore
                yield dataset_file

        elif self._fileobj is not None:
            try:
                yield self._fileobj
            finally:
                self._fileobj.seek(0)

        else:
            raise ValueError("Cannot return file object -- no filepath or fileobj available.")

    def __len__(self) -> int:
        if self._num_rows is None:
            self._num_rows = self.count_rows()
        return self._num_rows

    def count_rows(self) -> int:
        count = 0
        for _ in self.read_streaming_values():
            count += 1
        return count

    def get_columns(self) -> List[str]:
        stream = self.read_streaming_records()
        return list(next(stream).keys())

    @abstractmethod
    def read_streaming_records(self) -> Generator[RecordType, None, None]:
        pass

    @abstractmethod
    def read_streaming_values(self) -> Generator[List[Any], None, None]:
        pass

    @abstractmethod
    def read_file_as_dataframe(self) -> pd.DataFrame:
        pass



import pathlib
from typing import Any, Optional, TypeVar, Union
import math

import numpy as np
import pandas as pd

try:
    import pyspark.sql

    pyspark_exists = True
except ImportError:
    pyspark_exists = False

from .dataset_source import (
    DatasetSource,
    FilepathDatasetSource,
    PandasDatasetSource,
)

dataset_source_types = (
    Union[str, pathlib.Path, pd.DataFrame]
    if not pyspark_exists
    else Union[str, pathlib.Path, pd.DataFrame, pyspark.sql.DataFrame]
)

DatasetSourceType = TypeVar("DatasetSourceType", bound=dataset_source_types)  # type: ignore


def init_dataset_source(
    dataset_source: DatasetSourceType, dataset_name: Optional[str] = None
) -> DatasetSource:
    if isinstance(dataset_source, pd.DataFrame):
        if dataset_name is None:
            raise ValueError("Must provide dataset name if uploading from a DataFrame")
        return PandasDatasetSource(df=dataset_source, dataset_name=dataset_name)
    elif isinstance(dataset_source, pathlib.Path):
        return FilepathDatasetSource(filepath=dataset_source, dataset_name=dataset_name)
    elif isinstance(dataset_source, str):
        return FilepathDatasetSource(
            filepath=pathlib.Path(dataset_source), dataset_name=dataset_name
        )
    elif pyspark_exists and isinstance(dataset_source, pyspark.sql.DataFrame):
        from .dataset_source import PySparkDatasetSource

        if dataset_name is None:
            raise ValueError("Must provide dataset name if uploading from a DataFrame")
        return PySparkDatasetSource(df=dataset_source, dataset_name=dataset_name)
    else:
        raise ValueError("Invalid dataset source provided")


def check_none(x: Any) -> bool:
    if isinstance(x, str):
        return x == "None" or x == "none" or x == "null" or x == "NULL"
    elif isinstance(x, float):
        return math.isnan(x)
    elif pd.isnull(x):
        return True
    else:
        return x is None


def check_not_none(x: Any) -> bool:
    return not check_none(x)
